import io
import json
import logging
import sys

import pandas as pd
import sqlalchemy
from sqlalchemy.pool import StaticPool

import nexinfosys
from nexinfosys.command_definitions import commands
from nexinfosys.command_descriptions import c_descriptions
from nexinfosys.command_field_definitions import command_fields
from nexinfosys.command_field_descriptions import cf_descriptions
from nexinfosys.command_generators.parser_field_examples import generic_field_examples, generic_field_syntax
from nexinfosys.command_generators.parser_field_parsers import string_to_ast, arith_boolean_expression
from nexinfosys.common.create_database import create_pg_database_engine, create_monet_database_engine
from nexinfosys.common.helper import strcmp, add_label_columns_to_dataframe, generate_json
from nexinfosys.ie_exports.back_to_nis_format import nis_format_spreadsheet
from nexinfosys.ie_exports.flows_graph import construct_flow_graph_2
from nexinfosys.ie_exports.geolayer import generate_geojson
from nexinfosys.ie_exports.json_export import export_model_to_json
from nexinfosys.ie_exports.processors_graph import construct_processors_graph_2
from nexinfosys.ie_exports.sdmx import get_dataset_metadata
from nexinfosys.ie_exports.xml_export import export_model_to_xml
from nexinfosys.ie_imports.data_sources.eurostat_bulk import Eurostat
from nexinfosys.ie_imports.data_sources.eurostats_comext import COMEXT
from nexinfosys.ie_imports.data_sources.fadn import FADN
from nexinfosys.ie_imports.data_sources.faostat import FAOSTAT
from nexinfosys.ie_imports.data_sources.oecd import OECD
from nexinfosys.model_services import State, get_case_study_registry_objects
from nexinfosys.models.musiasem_concepts import Parameter, Hierarchy, ProblemStatement
from nexinfosys.models.musiasem_methodology_support import load_table, DBSession, User, Authenticator, CaseStudyStatus, \
    ObjectType, PermissionType, ORMBase
from nexinfosys.solving import BasicQuery

tm_permissions = {  # PermissionType
    "f19ad19f-0a74-44e8-bd4e-4762404a35aa": "read",
    "04cac7ca-a90b-4d12-a966-d8b0c77fca70": "annotate",
    "d0924822-32fa-4456-8143-0fd48da33fd7": "contribute",
    "83d837ab-01b2-4260-821b-8c4a3c52e9ab": "share",
    "d3137471-84a0-4bcf-8dd8-16387ea46a30": "delete"
}

tm_default_users = {  # Users
    "f3848599-4aa3-4964-b7e1-415d478560be": "admin",
    "2a512320-cef7-41c6-a141-8380d900761b": "_anonymous"
}

tm_object_types = {  # ObjectType
    "d6649a54-3538-4ee6-a4fc-8a67b74ed21f": "processor",
    "7eca7cfb-3eea-475f-9950-2d1093099ccb": "flow/fund",
    "3cc2582b-5142-4c8f-b484-a4693ba267cf": "flow",
    "c18ca5ab-471d-4ab4-9c0d-9563078a1c18": "fund",
    "5c9ef31b-399f-4007-824f-f2251e29bfdc": "ff_in_processor",
    "6e187787-adf2-4bdf-a0af-ab2801a6be42": "hierarchy",
    "91a7e3c2-115d-4166-a893-db6a19224154": "pedigree_matrix",
    "659f95f3-bc48-47e5-8584-64bd4477f3f2": "case_study"
}

tm_authenticators = {  # Authenticator
    "b33193c3-63b9-49f7-b888-ceba619d2812": "google",
    "c09fa36b-62a3-4904-9600-e3bb5028d809": "facebook",
    "f510cb30-7a44-4cb1-86f5-1b112e43293a": "firebase",
    "5f32a593-306f-4b69-983c-0a5680556fae": "local",
}

tm_case_study_version_statuses = {  # CaseStudyStatus
    "ee436cb7-0237-4b40-bd77-25acfade0f9b": "in_elaboration",
    "eef5c756-56ac-47a0-96cf-b7bebd73d392": "finished",
    "7d2cd2ae-0f4e-4962-b73c-4e40c553f533": "finished_and_published",
    "e5f3d8e6-accf-4175-9e74-5ad1b9a6faf5": "stopped"
}
# "22bc2577-883a-4408-bba9-fcace20c0fc8":
# "e80a7d27-3ec8-4aa1-b49c-5498e0f85bee":
# "d30120f0-28df-4bca-90e4-4f0676d1c874":
# "83084df6-7ad0-45d7-b3f1-6de594c78611":
# "7e23991b-24a0-4da1-8251-c3c3434dfb87":
# "bfc0c9fe-631f-44d0-8e96-e22d01ffb1ed":
# "dec7e901-b3f4-4343-b3d1-4fa5fbf3804e":
# "013b2f3b-4b2f-4b6c-8f5f-425132aea74b":
# "3eef41be-fde1-4ad4-92d0-fe795158b41d":
# "0fba3591-4ffc-4a88-977a-6e1d922f0735":
# "a61fc587-1272-4d46-bdd0-027cde1b8a78":
# "600397ef-0102-486e-a6f7-d43b0f8ce4b9":
# "763e57b7-2636-4c04-9861-d865fe0bb5ab":
# "788065a7-d9f5-46fa-b8ba-8bc223d09331":
# "38fb34f7-a952-4036-9b0b-4d6c59e8f8d4":
# "0292821a-dd33-450a-bdd8-813b2b95c456":


def initialize_database_data():
    # Load base tables
    load_table(DBSession, User, tm_default_users)
    load_table(DBSession, Authenticator, tm_authenticators)
    load_table(DBSession, CaseStudyStatus, tm_case_study_version_statuses)
    load_table(DBSession, ObjectType, tm_object_types)
    load_table(DBSession, PermissionType, tm_permissions)
    # Create and insert a user
    session = DBSession()
    # Create test User, if it does not exist
    u = session.query(User).filter(User.name == 'test_user').first()
    if not u:
        u = User()
        u.name = "test_user"
        u.uuid = "27c6a285-dd80-44d3-9493-3e390092d301"
        session.add(u)
        session.commit()
    DBSession.remove()


def initialize_databases():
    recreate_db = False
    if nexinfosys.get_global_configuration_variable("DB_CONNECTION_STRING"):
        db_connection_string = nexinfosys.get_global_configuration_variable("DB_CONNECTION_STRING")
        echo_sql = nexinfosys.get_global_configuration_variable("SQL_ECHO", False)
        if isinstance(echo_sql, str):
            echo_sql = True if echo_sql.lower() == "true" else False
        logging.debug("Connecting to metadata server")
        logging.debug(db_connection_string)
        logging.debug("-----------------------------")
        if db_connection_string.startswith("sqlite://"):
            nexinfosys.engine = sqlalchemy.create_engine(db_connection_string,
                                                         echo=echo_sql,
                                                         connect_args={'check_same_thread': False},
                                                         poolclass=StaticPool)
        else:
            nexinfosys.engine = create_pg_database_engine(db_connection_string, "magic_nis", recreate_db=recreate_db)

        # global DBSession # global DBSession registry to get the scoped_session
        DBSession.configure(bind=nexinfosys.engine)  # reconfigure the sessionmaker used by this scoped_session
        tables = ORMBase.metadata.tables
        connection = nexinfosys.engine.connect()
        table_existence = [nexinfosys.engine.dialect.has_table(connection, tables[t].name) for t in tables]
        connection.close()
        if False in table_existence:
            ORMBase.metadata.bind = nexinfosys.engine
            ORMBase.metadata.create_all()
        # connection = nexinfosys.engine.connect()
        # table_existence = [nexinfosys.engine.dialect.has_table(connection, tables[t].name) for t in tables]
        # connection.close()
        # Load base tables
        initialize_database_data()
    else:
        logging.error("No database connection defined (DB_CONNECTION_STRING), exiting now!")
        sys.exit(1)

    if nexinfosys.get_global_configuration_variable("DATA_CONNECTION_STRING"):
        data_connection_string = nexinfosys.get_global_configuration_variable("DATA_CONNECTION_STRING")
        logging.debug("Connecting to data server")
        if data_connection_string.startswith("monetdb"):
            # TODO Install monet packages (see commented packages in "requirements.txt")
            nexinfosys.data_engine = create_monet_database_engine(data_connection_string, "magic_data")
        elif data_connection_string.startswith("sqlite://"):
            nexinfosys.data_engine = sqlalchemy.create_engine(data_connection_string,
                                                              echo=echo_sql,
                                                              connect_args={'check_same_thread': False},
                                                              poolclass=StaticPool)
        else:
            nexinfosys.data_engine = create_pg_database_engine(data_connection_string, "magic_data", recreate_db=recreate_db)
    else:
        logging.error("No data connection defined (DATA_CONNECTION_STRING), exiting now!")
        sys.exit(1)


def get_parameters_in_state(state: State):
    res = []
    registry, _, _, _, _ = get_case_study_registry_objects(state)
    for p in registry.get(Parameter.partial_key()):
        p_name = p.name
        p_type = p.type
        if p.range:
            if strcmp(p_type, "Number"):
                p_range = p.range
            else:
                glb_idx, _, _, _, _ = get_case_study_registry_objects(state)
                h = glb_idx.get(Hierarchy.partial_key(p.range))
                h = h[0]
                p_range = ', '.join(h.codes.keys())
        else:
            p_range = ""
        res.append(dict(name=p_name, type=p_type, range=p_range))
    return res


def get_scenarios_in_state(state: State):
    glb_idx, _, _, _, _ = get_case_study_registry_objects(state)

    ps = glb_idx.get(ProblemStatement.partial_key())
    if len(ps) == 0:
        ps = [ProblemStatement()]

    scenarios = []
    for scenario, params in ps[0].scenarios.items():
        scenarios.append(dict(name=scenario, parameters=params.items()))

    return scenarios


def register_external_datasources():
    from nexinfosys.ie_imports.data_source_manager import DataSourceManager
    dsm2 = DataSourceManager(session_factory=DBSession)

    # Eurostat
    dsm2.register_datasource_manager(Eurostat())

    # TODO COMEXT
    dsm2.register_datasource_manager(COMEXT())

    # FAO
    fao_dir = nexinfosys.get_global_configuration_variable("FAO_DATASETS_DIR", "/home/rnebot/DATOS/FAOSTAT/")
    dsm2.register_datasource_manager(FAOSTAT(datasets_directory=fao_dir,
                                             metadata_session_factory=DBSession,
                                             data_engine=nexinfosys.data_engine))

    # OECD
    dsm2.register_datasource_manager(OECD())

    # FADN
    dsm2.register_datasource_manager(FADN(metadata_session_factory=DBSession,
                                          data_engine=nexinfosys.data_engine))
    # sources = dsm2.get_supported_sources()
    return dsm2


def get_graph_from_state(state: State, name: str):
    def html_template():
        return """
        <!doctype html>
        <html>
        <head>
          <title>WDS - VisJS Network</title>
          <script type="text/javascript" src="https://unpkg.com/vis-network/standalone/umd/vis-network.min.js"></script>

          <style type="text/css">
            #mynetwork {
              width: 1200px;
              height: 800px;
              border: 1px solid lightgray;
            }
          </style>
        </head>
        <body>

        <p>
          %s
        </p>

        <div id="mynetwork"></div>

        <script type="text/javascript">
          // create an array with nodes
          var nodes = new vis.DataSet(%s);
        // create an array with edges
          var edges = new vis.DataSet(%s);

          // create a network
          var container = document.getElementById('mynetwork');
          var data = {
            nodes: nodes,
            edges: edges
          };
          var options = {};
          var network = new vis.Network(container, data, options);
        </script>

        </body>
        </html>
    """

    if "." in name:
        pos = name.find(".")
        extension = name[pos + 1:]
        name = name[:pos]

    output = None
    mimetype = None
    if name == "interfaces_graph":
        if extension in ("visjs", "html"):
            query = BasicQuery(state)
            output = construct_flow_graph_2(state, query, None, "visjs")  # Version 2 !!!

            if extension == "html":
                output = html_template() % ("", output["nodes"], output["edges"])
                mimetype = "application/html"
            else:
                output = generate_json(output)
                mimetype = "application/json"
        elif extension == "gml":
            # Prepare GML file
            query = BasicQuery(state)
            output = construct_flow_graph_2(state, query, None, extension)  # Version 2 !!!
            mimetype = "application/text"  # TODO
    elif name == "processors_graph":
        if extension in ("visjs", "html"):
            query = BasicQuery(state)
            output = construct_processors_graph_2(state, query, None, True, True, False, "visjs")
            if extension == "html":
                output = html_template() % ("", output["nodes"], output["edges"])
                mimetype = "application/html"
            else:
                output = generate_json(output)
                mimetype = "application/json"
        elif extension == "gml":
            query = BasicQuery(state)
            output = construct_processors_graph_2(state, query, None, True, True, False, extension)
            mimetype = "application/text"

    return output, mimetype, output is not None


def get_dataset_from_state(state: State, name: str, extension: str, labels_enabled: bool):
    glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)
    if "." in name:
        pos = name.find(".")
        extension = name[pos + 1:]
        name = name[:pos]
    if name in datasets:
        # Obtain the data and the metadata
        ds = datasets[name]  # type: Dataset
        ds2 = ds.data
        # Labels
        if labels_enabled:
            logging.debug("Preparing Dataset labels")
            ds2 = add_label_columns_to_dataframe(name, ds2, glb_idx)

        if isinstance(ds2.index, (pd.Int64Index, pd.core.indexes.range.RangeIndex)):
            export_index = False
        else:
            export_index = True

        # TODO Elaborate "meta-workbook" (NIS workbook capable of reproducing the dataset)
        if extension == "json":
            tmp = json.loads(
                '{"data": ' + ds2.to_json(orient='split', date_format='iso', date_unit='s') + ', "metadata": {}}')
            del tmp["data"]["index"]
            return tmp, "text/json", True
        elif extension == "json2":
            tmp = ds2.to_json(orient='records', date_format='iso', date_unit='s')
            return tmp, "text/json", True
        elif extension == "csv":
            tmp = ds2.to_csv(date_format="%Y-%m-%d %H:%M:%S", index=export_index, na_rep="")
            return io.StringIO(tmp), "text/csv", True
        elif extension == "kendopivotgrid":
            # Prepare Data and Schema in Kendo PivotGrid format
            data = json.loads(ds2.to_json(orient='records', date_format='iso', date_unit='s'))
            fields = {}
            dimensions = {}
            measures = {}
            for i, c in enumerate(ds2.columns):
                is_dimension = i < len(ds2.columns) - 1
                fields[c] = dict(type="string" if is_dimension else "number")
                if is_dimension:
                    dimensions[c] = dict(caption="All " + c)
                else:
                    measures["Sum " + c] = dict(field=c, format="{0:c}", aggregate="sum")
                    measures["Avg " + c] = dict(field=c, format="{0:c}", aggregate="average")
            schema = dict(model=dict(fields=fields), cube=dict(dimensions=dimensions, measures=measures))
            return dict(data=data, schema=schema), "text/json", True
        elif extension == "sdmx.json":
            md = get_dataset_metadata(name, ds)
            return md, "text/json", True
        elif extension == "xlsx":
            # Generate XLSX from data & return it
            output = io.BytesIO()
            # from pyexcelerate import Workbook, Writer
            # wb = Workbook()
            # data = [ds2.columns] + [ds2.values]
            # wb.new_sheet(name, data=data)
            # wr = Writer.Writer(wb)
            # wr.save(output)
            logging.debug("Generating Excel")
            ds2.to_excel(output, sheet_name=name, index=False)  # , engine="xlsxwriter")
            return output.getvalue(), "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet", True
        else:
            return ds2, "application/pandas.dataframe", True
    else:
        return {"error": f"Could not find a Dataset with name '{name}' in the current state"}, "text/json", False


def get_model(state: State, format: str):
    output = None
    # Generate graph from State
    if state:
        if format == "json":
            output = str(export_model_to_json(state))
            mimetype = "text/json"
        elif format == "xlsx":
            # A reproducible session must be open, signal about it if not
            output = nis_format_spreadsheet(state)
            mimetype = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
        elif format == "xml":
            # A reproducible session must be open, signal about it if not
            # Prepare a XML string
            glb_idx, _, _, _, _ = get_case_study_registry_objects(state)
            output, _ = export_model_to_xml(glb_idx)
            mimetype = "text/xml"

    if output:
        return output, mimetype, True
    else:
        return {"error": F"Cannot return model, format '{format}' not recognized"}, "text/json", False


def get_geolayer(state: State, format):
    output = None
    # Generate graph from State
    if state:
        # TODO Obtain a list of Geolocated processors
        #      Obtain attributes attached to the processor: Name, Full name, system, subsystem type, ...
        #      Go to general matrix and obtain values for interfaces, different scenarios, times and observers
        # TODO Extract geometry of processors into a new layer
        # TODO  Elaborate a new layer where each processor
        if format == "geojson":
            # TODO Prepare GeoJSON file
            output = generate_geojson(state)
            output = io.StringIO(generate_json(output))
            mimetype = "application/geo+json"
        elif format == "kmz" or format == "kml":
            # TODO Prepare KML file
            output = io.BytesIO()
            if format == "kmz":
                mimetype = "application/vnd.google-earth.kmz"
            else:
                mimetype = "application/vnd.google-earth.kml+xml"

    if output:
        return output.getvalue(), mimetype, True
    else:
        return {"error": F"Cannot return geolayer, format '{format}' not recognized"}, "text/json", False


def get_ontology(state: State, format):
    # TODO OWLREADY2 installation on the Docker image issues a problem
    # Recover InteractiveSession
    output = None
    # Generate graph from State
    if state:
        if format == "owl":
            # TODO Prepare OWL file
            output = io.StringIO()
            mimetype = "application/rdf+xml"  # TODO

    if output:
        return output.getvalue(), mimetype, True
    else:
        return {"error": F"Cannot return ontology, format '{format}' not recognized"}, "text/json", False


def validate_command(command_content_to_validate):
    """
    The input comes in a JSON field "content":
    {"command": "<command name",
     "fields": {"<field name>": "<value", ...}
    }
    :param command_content_to_validate:
    :return: A dictionary with the same fields of the input dictionary, whose values are the diagnosis, None being
            everything-ok, and a string being a message describing the problem.
    """
    def split_expansion_expressions(f, content):
        # Dataset expansion. Isolate each expression
        pieces = []
        offset = 0
        look_for = "{"
        open_brace = False
        s = None
        while offset < len(content):
            pos = content[offset:].find(look_for)
            if pos >= 0:
                if look_for == "{":
                    if pos > 0:
                        pieces.append((content[offset:offset + pos], False))  # Literal
                    look_for = "}"
                    open_brace = True
                else:
                    if pos > 0:
                        pieces.append((content[offset:offset + pos], True))  # Expansion
                    else:
                        s = f"Invalid syntax in field '{f}' with value: " + content + ". No expression between curly braces."
                        valid = False
                        break
                    look_for = "{"
                    open_brace = False
                offset += pos + 1
            else:  # Character not found
                if open_brace:
                    s = f"Invalid syntax in field '{f}' with value: " + content + ". Curly brace not closed."
                    valid = False
                    break
                else:  # Add the rest
                    pieces.append((content[offset:], False))
                    offset = len(content)

        return pieces, s

    if "command" in command_content_to_validate:
        command = command_content_to_validate["command"]
    else:
        raise Exception("Must specify 'command'")

    if "fields" in command_content_to_validate:
        fields = command_content_to_validate["fields"]
    else:
        raise Exception("Must specify 'fields'")

    alternative_command_names = command_content_to_validate.get("alternative_command_names", {})

    result = {}
    # Find command from the worksheet name ("command")
    match = None
    for cmd in commands:
        for cmd_name in cmd.allowed_names:
            if cmd_name.lower() in command.lower():
                if match:
                    if match[1] < len(cmd_name):
                        match = (cmd.name, len(cmd_name))
                else:
                    match = (cmd.name, len(cmd_name))
    if not match:
        for k, v in alternative_command_names:
            if k.lower() in command.lower():
                for cmd in commands:
                    for cmd_name in cmd.allowed_names:
                        if cmd_name.lower() in v.lower():
                            match = (cmd.name, 0)
                            break
                    if match:
                        break
                if match:
                    break

    # Fields in the command
    status = True
    if match:
        for f in fields:  # Validate field by field
            for f2 in command_fields[match[0]]:  # Find corresponding field in the command
                if f.lower() in [f3.lower() for f3 in f2.allowed_names]:
                    fld = f2
                    break
            else:
                fld = None
            if fld:  # If found, can validate syntax
                # Validate Syntax
                content = fields[f]
                content_msg = content  # Original "content", to show in case of error
                if isinstance(content, (int, float)):
                    content = str(content)

                # Check if it is an expansion expression
                valid = True
                if "{" in content or "}" in content:
                    # Is expansion allowed in this command?
                    expansion_allowed = True
                    if expansion_allowed:
                        pieces, s = split_expansion_expressions(f, content)
                        if s is None:
                            c = ""
                            for p in pieces:
                                if p[1]:  # Expansion expression
                                    try:
                                        string_to_ast(arith_boolean_expression, p[0])
                                        c += "expand"
                                    except:
                                        s = f"Invalid syntax in field '{f}' with value: {content}, expansion expression '{p[0]}' invalid"
                                        result[f] = s
                                        valid = False
                                        break
                                else:
                                    c += p[0]
                            if valid:
                                content = c
                        else:
                            valid = False

                if not valid:
                    result[f] = s
                    status = False
                else:
                    if fld.allowed_values:
                        if content != content_msg:  # It was an expansion expression, cannot check it now, assume it is good
                            result[f] = None
                        else:
                            # Case insensitive comparison
                            if content.lower().strip() in [f.lower().strip() for f in fld.allowed_values]:
                                result[f] = None
                            else:
                                result[f] = "'"+content+"' in field '"+f+"' must be one of: "+", ".join(fld.allowed_values)
                                status = False
                    else:
                        try:
                            string_to_ast(fld.parser, content)
                            result[f] = None
                        except:
                            s = f"Invalid syntax in field '{f}' with value: '{content_msg}'"
                            if fld.examples:
                                s += ". Examples: "+", ".join(fld.examples)
                            result[f] = s
                            status = False

            else:
                result[f] = "Field '"+f+"' not found in command '"+command+"'. Possible field names: "+", ".join([item for f2 in command_fields[command] for item in f2.allowed_names])
                status = False
    else:
        for f in fields:  # Validate field by field
            result[f] = "Command '" + command +"' not found in the list of command names: " +", ".join([n for c in commands for n in c.allowed_names])
        status = False

    return result, status


def command_field_help(command_content_to_validate):
    """
    The input comes in a JSON field "content":
    {"command": "<command name",
     "fields": ["<field name>", "<field_name>"]
    }

    :return: A dictionary with the same fields passed in the input dictionary, whose values are the help divided in
        sections: explanation, allowed_values, formal syntax and examples
    """
    if "command" in command_content_to_validate:
        command = command_content_to_validate["command"]
    else:
        raise Exception("Must specify 'command'")

    if "fields" in command_content_to_validate:
        fields = command_content_to_validate["fields"]
    else:
        raise Exception("Must specify 'fields'")

    alternative_command_names = command_content_to_validate.get("alternative_command_names", {})

    result = {}
    # Find command from the worksheet name ("command")
    match = None
    for cmd in commands:
        for cmd_name in cmd.allowed_names:
            if cmd_name.lower() in command.lower():
                if match:
                    if match[1] < len(cmd_name):
                        match = (cmd.name, len(cmd_name))
                else:
                    match = (cmd.name, len(cmd_name))
    if not match:
        for k, v in alternative_command_names:
            if k.lower() in command.lower():
                for cmd in commands:
                    for cmd_name in cmd.allowed_names:
                        if cmd_name.lower() in v.lower():
                            match = (cmd.name, 0)
                            break
                    if match:
                        break
                if match:
                    break

    # Fields in the command
    status = True
    if match:
        for f in fields:  # Validate field by field
            for f2 in command_fields.get(match[0], []):  # Find corresponding field in the command
                if f.lower() in [f3.lower() for f3 in f2.allowed_names]:
                    fld = f2
                    break
            else:
                fld = None
            if fld:  # If found, can elaborate help
                mandatoriness = "Mandatory" if fld.mandatory else "Optional"
                description = cf_descriptions.get((match[0], fld.name), f"Text for field {f} in command {match[0]}")
                if isinstance(description, list):
                    description = description[0]  # Short version for online. Index 1 contains the long version
                examples = ("<br><b>Examples:</b><br>&nbsp;&nbsp;"+"<br>&nbsp;&nbsp;".join(generic_field_examples[fld.parser])) if generic_field_examples.get(fld.parser) else ""
                syntax = ", ".join(fld.allowed_values) if fld.allowed_values else generic_field_syntax.get(fld.parser, "<>")
                if fld.allowed_values:
                    result[f] = f"<small><b>({mandatoriness})</b></small><br>{description}<br><b>Syntax. One of:</b> {syntax}"
                else:
                    result[f] = f"<small><b>({mandatoriness})</b></small><br>{description}<br><b>Syntax:</b> {syntax}{examples}"
            else:
                if match[0] == "datasetqry" and f:  # Must be a dimension name
                    mandatoriness = "Optional"
                    description = "A dimension name. Use for filtering purposes (leave the values that MUST be at the output)."
                    result[f] = f"<small><b>({mandatoriness})</b></small><br>{description}"
                else:
                    result[f] = "Field '"+f+"' not found in command '"+command+"'. Possible field names: "+", ".join([item for f2 in command_fields.get(match[0], []) for item in f2.allowed_names])
                    status = False
    else:
        for f in fields:  # Validate field by field
            result[f] = "Command '" + command +"' not found in the list of command names: " +", ".join([n for c in commands for n in c.allowed_names])
        status = False

    return result, status


def comm_help(command_content_to_validate):
    if "command" in command_content_to_validate:
        command = command_content_to_validate["command"]
    else:
        raise Exception("Must specify 'command'")

    alternative_command_names = command_content_to_validate.get("alternative_command_names", {})

    # Find command from the worksheet name ("command")
    match = None
    for cmd in commands:
        for cmd_name in cmd.allowed_names:
            if cmd_name.lower() in command.lower():
                if match:
                    if match[1] < len(cmd_name):
                        match = (cmd.name, len(cmd_name))
                else:
                    match = (cmd.name, len(cmd_name))
    if not match:
        for k, v in alternative_command_names:
            if k.lower() in command.lower():
                for cmd in commands:
                    for cmd_name in cmd.allowed_names:
                        if cmd_name.lower() in v.lower():
                            match = (cmd.name, 0)
                            break
                    if match:
                        break
                if match:
                    break

    # Fields in the command
    status = True
    if match:
        if (match[0], "title") in c_descriptions:
            title = c_descriptions[(match[0], "title")]
            description = c_descriptions[(match[0], "description")]
            semantics = c_descriptions[(match[0], "semantics")]
            examples = c_descriptions[(match[0], "examples")]
            result = dict(title=title, description=description, semantics=semantics, examples=examples)
        else:
            result = dict(title=f"Title {command}", description=f"Description {command}", semantics=f"Semantics {command}")
    else:
        status = False
        result = dict()

    return result, status


def prepare_and_reset_database_for_tests(prepare=False, metadata_string="sqlite://", data_string="sqlite://"):
    if prepare:
        nexinfosys.engine = sqlalchemy.create_engine(metadata_string, echo=True)
        nexinfosys.data_engine = sqlalchemy.create_engine(data_string, echo=True)

        # global DBSession # global DBSession registry to get the scoped_session
        DBSession.configure(bind=nexinfosys.engine)  # reconfigure the sessionmaker used by this scoped_session
        tables = ORMBase.metadata.tables
        connection = nexinfosys.engine.connect()
        table_existence = [nexinfosys.engine.dialect.has_table(connection, tables[t].name) for t in tables]
        connection.close()
        if False in table_existence:
            ORMBase.metadata.bind = nexinfosys.engine
            ORMBase.metadata.create_all()

    # Load base tables
    load_table(DBSession, User, tm_default_users)
    load_table(DBSession, Authenticator, tm_authenticators)
    load_table(DBSession, CaseStudyStatus, tm_case_study_version_statuses)
    load_table(DBSession, ObjectType, tm_object_types)
    load_table(DBSession, PermissionType, tm_permissions)
    # Create and insert a user
    session = DBSession()
    # Create test User, if it does not exist
    u = session.query(User).filter(User.name == 'test_user').first()
    if not u:
        u = User()
        u.name = "test_user"
        u.uuid = "27c6a285-dd80-44d3-9493-3e390092d301"
        session.add(u)
        session.commit()
    DBSession.remove()


import base64
import json
import uuid
from abc import ABCMeta, abstractmethod  # Abstract Base Class

import logging
from typing import List, Union, Dict, Any

from nexinfosys import Issue, IssuesOutputPairType
from nexinfosys.common.helper import create_dictionary, PartialRetrievalDictionary

logger = logging.getLogger(__name__)


class IExecutableCommand(metaclass=ABCMeta):
    """ A command prepared for its execution. Commands have direct access to the current STATE """

    @abstractmethod
    def execute(self, state: "State") -> IssuesOutputPairType:
        """
        Execute the command. At the same time, generate a list of issues.
        At this point, it is assumed there are no syntactic errors

        :param state:
        :return: (list of issues, output)
        """
        raise Exception("Execute not implemented")
        # return None, None  # Issues, Output

    @abstractmethod
    def estimate_execution_time(self):
        pass

    @abstractmethod
    def json_serialize(self) -> Dict:
        pass

    # @abstractmethod
    def json_deserialize(self, json_input: Union[dict, str, bytes, bytearray]) -> List[Issue]:
        """
        Read command parameters from a JSON string
        Check the validity of the JSON input
        After this, the command can be executed ("execute")
        
        :param json_input: JSON in a Unicode String 
        :return: --- (the object state is updated, ready for execution)
        """
        issues = []
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)
        return issues


class Scope:
    """ The scope allows to assign names to entities using a registry """

    def __init__(self, name=None):
        self._name = name  # A name for the scope itself
        self._registry = create_dictionary()

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    def __contains__(self, key):  # "in" operator to check if the key is present in the dictionary
        return key in self._registry

    def __getitem__(self, name):
        if name in self._registry:
            return self._registry[name]
        else:
            return None

    def __setitem__(self, name: str, entity):
        if name not in self._registry:
            existing = True
        else:
            existing = False

        self._registry[name] = entity
        return existing

    def __delitem__(self, name):
        del self._registry[name]

    def list(self):
        """ List just the names of variables """
        return self._registry.keys()

    def list_pairs(self):
        """ List tuples of variable name and value object """
        return [(k, v2) for k, v2 in self._registry.items()]


class Namespace:
    def __init__(self):
        self.__scope = []  # List of scopes
        self.__current_scope = None  # type: Scope
        self.__current_scope_idx = -1
        self.new_scope()

    # The registry will have "nested" scopes (names in the current scope take precedence on "higher" scopes)
    # When searching for names, the search will go from the most recent scope to the oldest
    def new_scope(self, name=None):
        """ Create a new scope """
        self.__current_scope = Scope()
        self.__scope.append(self.__current_scope)
        self.__current_scope_idx = len(self.__scope) - 1
        if not name:
            name = "Scope" + str(self.__current_scope_idx)
        self.__current_scope.name = name

    def close_scope(self):
        if self.__current_scope:
            del self.__scope[-1]
        if self.__current_scope_idx >= 0:
            self.__current_scope_idx -= 1
            if self.__current_scope_idx >= 0:
                self.__current_scope = self.__scope[-1]
            else:
                self.__current_scope = None

    def list_names(self, scope=None):
        """ Returns a list of the names of the registered entities of the "scope" or if None, of the CURRENT scope """
        if not scope:
            scope = self.__current_scope

        return scope.list()

    def list(self, scope=None):
        """
            Returns a list of the names and values of the registered entities of
            the "scope" or if None, of the CURRENT scope
        """
        if not scope:
            scope = self.__current_scope

        return scope.list_pairs()

    def list_all_names(self):
        """
            Returns a list of the names of registered entities considering the scopes
            Start from top level, end in bottom level (the current one, which takes precedence)
            :return:
        """
        t = create_dictionary()
        for scope in self.__scope:
            t.update(scope._registry)

        return t.keys()

    def list_all(self):
        """
            Returns a list of the names and variables of registered entities considering the scopes
            Start from top level, end in bottom level (the current one, which takes precedence)

        :return:
        """
        t = create_dictionary()
        for scope in self.__scope:
            t.update(scope._registry)

        return [(k, v2) for k, v2 in t.items()]

    def set(self, name: str, entity):
        """ Set a named entity in the current scope. Previous scopes are not writable. """
        if self.__current_scope:
            var_exists = name in self.__current_scope
            self.__current_scope[name] = entity
            # if var_exists:
            #     logger.warning("'" + name + "' overwritten.")

    def get(self, name: str, scope=None, return_scope=False):
        """ Return the entity named "name". Return also the Scope in which it was found """
        if not scope:
            for scope_idx in range(len(self.__scope) - 1, -1, -1):
                if name in self.__scope[scope_idx]:
                    if return_scope:
                        return self.__scope[scope_idx][name], self.__scope[scope_idx]
                    else:
                        return self.__scope[scope_idx][name]
            else:
                # logger.warning(
                #     "The name '" + name + "' was not found in the stack of scopes (" + str(len(self.__scope)) + ")")
                if return_scope:
                    return None, None
                else:
                    return None
        else:
            # TODO Needs proper implementation !!!! (when scope is a string, not a Scope instance, to be searched in the list of scopes "self.__scope")
            if name in scope:
                if return_scope:
                    return scope[name], scope
                else:
                    return scope[name]
            else:
                logger.error("The name '" + name + "' was not found in scope '" + scope.name + "'")
                if return_scope:
                    return None, None
                else:
                    return None


class State:
    """
    -- "State" in memory --

    Commands may alter State or may just read it
    It uses a dictionary of named Namespaces (and Namespaces can have several scopes)
    Keeps a registry of variable names and the objects behind them.

        It is basically a list of Namespaces. One is active by default.
        The others have a name. Variables inside these other Namespaces may be accessed using that
        name then "::", same as C++

    State Serialization functions specialized in the way State is used in MuSIASEM are in the "serialization" module:

    serialize_state
    deserialize_state

    """

    def __init__(self, d: Dict[str, Any] = None):
        self._default_namespace = ""
        self._namespaces = create_dictionary()

        if d is not None and len(d) > 0:
            self.update(d)

    def new_namespace(self, name):
        self._namespaces[name] = Namespace()
        if self._default_namespace is None:
            self._default_namespace = name

    @property
    def default_namespace(self):
        return self._default_namespace

    @default_namespace.setter
    def default_namespace(self, name):
        if name is not None:  # Name has to have some value
            self._default_namespace = name

    def del_namespace(self, name):
        if name in self._namespaces:
            del self._namespaces

    def list_namespaces(self):
        return self._namespaces.keys()

    def list_namespace_variables(self, namespace_name=None):
        if namespace_name is None:
            namespace_name = self._default_namespace

        return self._namespaces[namespace_name].list_all()

    def update(self, d: Dict[str, Any], namespace_name=None):
        if namespace_name is None:
            namespace_name = self._default_namespace

        if namespace_name not in self._namespaces:
            self.new_namespace(namespace_name)

        for name, entity in d.items():
            self._namespaces[namespace_name].set(name, entity)
        # self._namespaces[namespace_name].update(d)

    def set(self, name, entity, namespace_name=None):
        if namespace_name is None:
            namespace_name = self._default_namespace

        if namespace_name not in self._namespaces:
            self.new_namespace(namespace_name)

        self._namespaces[namespace_name].set(name, entity)

    def get(self, name, namespace_name=None, scope=None):
        if not namespace_name:
            namespace_name = self._default_namespace

        if namespace_name not in self._namespaces:
            self.new_namespace(namespace_name)

        return self._namespaces[namespace_name].get(name, scope)


def get_case_study_registry_objects(state, namespace=None):
    """
    Obtain the main entries of the state

    :param state: Input state (modified also)
    :param namespace: State supports several namespaces. This one serves to specify which one. Default=None
    :return: Tuple: (global index, processor sets, hierarchies, datasets, mappings)
    """
    # Index of ALL objects
    glb_idx = state.get("_glb_idx", namespace)
    if not glb_idx:
        glb_idx = PartialRetrievalDictionary()
        state.set("_glb_idx", glb_idx, namespace)

    # ProcessorSet dict (dict of sets)
    p_sets = state.get("_processor_sets", namespace)
    if not p_sets:
        p_sets = create_dictionary()
        state.set("_processor_sets", p_sets, namespace)

    # Hierarchies Dict
    hh = state.get("_hierarchies", namespace)
    if not hh:
        hh = create_dictionary()
        state.set("_hierarchies", hh, namespace)
    # Datasets Dict
    datasets = state.get("_datasets", namespace)
    if not datasets:
        datasets = create_dictionary()
        state.set("_datasets", datasets, namespace)
    # Mappings Dict
    mappings = state.get("_mappings", namespace)
    if not mappings:
        mappings = create_dictionary()
        state.set("_mappings", mappings, namespace)

    return glb_idx, p_sets, hh, datasets, mappings


class LocallyUniqueIDManager:
    """
    Obtains UUID but encoded in base85, which still is ASCII, but is more compact than the UUID standard hexadecimal
    representation
    """
    class __LocallyUniqueIDManager:
        def __init__(self, c: int = 0):
            self.val = c

        def get_new_id(self, inc):
            return base64.a85encode(uuid.uuid1().bytes).decode("ascii")

        def __str__(self):
            return repr(self) + self.val
    instance = None

    def __init__(self, arg=0):
        if not LocallyUniqueIDManager.instance:
            LocallyUniqueIDManager.instance = LocallyUniqueIDManager.__LocallyUniqueIDManager(arg)
        else:
            LocallyUniqueIDManager.instance.val = arg

    def get_new_id(self, inc: int = 1):
        return self.instance.get_new_id(inc)

    def __getattr__(self, name):
        return getattr(self.instance, name)


"""

API
* Open/close interactive session
* Identify
* Open a reproducible session (optionally load existing command_executors/state). Close it, optionally save.
* CRUD case studies, versions and variables (objects)

* Browse datasets
* Browse case study objects: mappings, hierarchies, grammars

* Submit Worksheet
  - Interactive session
  - Open work session
  - Submit file
	- the file produces a sequence of command_executors
	- execute
	  - elaborate output file and compile issues
  - Close Interactive session (save, new case study version)
  - Close user session
* Submit R script
  - Interactive session
  - Open work session
  - Submit file
	- the file produces command_executors
	- execute
	  - elaborate output file ¿?, compile issues
  - Close work session (save, new case study version)
  - Close Interactive session
* Execute sparse command_executors (from R client, Python client, Google Sheets...)
  - Interactive session
  - parse then execute the command

* Export/import
  - Internal export. No need to be identified
  - Packaged case study export into Zenodo
  - Import case study (reverse of "internal export"). No need to be identified, the import needs a close work session with save or not
* Manage users
  - Link authenticators with identities
* Manage permissions
  - Share/unshare, READ, CONTRIBUTE, SHARE, EXPORT

RESTFul

* Open/close the interactive session
  - Open interactive, persisted session. Creation date registered. Register from where does the connection come
    (last call is also registered so after some time the session can be closed)
    POST /isession  -> return the ID, generate a COOKIE for the interactive session
  - Close interactive session
    DELETE /isession/ -> (the cookie has to be passed)
* Identify
  - Method 1. Through a OAuth2 token

USE CASES
* Start a case study from scratch
* Continue working on a case study. Implies saving the case study. A status flag for the case study (in elaboration, ready, publishable, ...)
* Play with an existing case study. CRUD elements, modify parameters, solve, analyze (read), ....
* Start an anonymous case study from scratch. It will not be saved
* Create case study from an existing one, as new version ("branch") or totally new case study
  *
* Analyze case study (open a case study, play with it)

"""

"""
Module containing high level function calls, controlling the different use cases
These functions are called by the RESTful interface, which is the gate for the different other clients: R client, Web,...

The functions allow a user to start a Work Session on a Case Study. All work is in memory. Specific functions allow
storing in database or in file.

A Work Session has a memory state related to one or more case studies, and a user. To shorten command_executors, there will be a
default case study.

* Queries (to database) encompassing several case studies? That would break the assumption of a case study per work
  session ¿do not assume that then? It implies having a default case study -context- to simplify command_executors
  * It would be like multi-database queries. The result is a valid structure, the retrieval process has to deal with opening
* Spin-off objects (objects born inside a case study which can be used in other case studies). Consider clone or
  reference options.


Use cases:
* Start a case study from scratch
* Continue working on a case study. Implies saving the case study. A status flag for the case study (in elaboration, ready, publishable, ...)
* Play with an existing case study. CRUD elements, modify parameters, solve, analyze (read), ....
* Start an anonymous case study from scratch. It will not be saved
* Create case study from an existing one, as new version ("branch") or totally new case study
  *
* Export case study for publication
* Export objects
* Analyze case study
* Import case study
* Import objects


Work Session level command types:
* Open a session
* Create a case study
* Load case study (from DB, from file)
* Clone case study. Specify if it is a new version or a new case study
  * If it is a new case study, reset metadata
* Save case study. Only for case studies with metadata.
  * The case study
  *
* Case study command_executors
  * Execute
  * Add to the case study


Case study level command types:
* Case study Metadata ("librarian/archivist")
* Data import ("traders"). Prepare external data for its use into the case study
* Specification ("writers"). Modify the structures and values composing the case study, including metadata of elements of the case study
* Solve ("info-reactors"). Deduce new information from existing structures/constraints
* Read/Query ("readers")
* Visualization ("story tellers")

(In-memory) Object types:
* Work session
  * Work session variables
* Case study
* Hierarchy
  * Categories
  * Flows
  * Funds
* Transformation from a hierarchy to another
* Indicator + Benchmark
  * Metabolic rate
  * Many others
* Grammar. A grammar is specialized in a case study. It also serves as a set of constraints.
  * Changes should be allowed before "instantiation"
* Sequences connection
* Upscale transform
* Dataset
* Data process
* Dataset source
* NUSAP matrix or profile
* Analysis matrices: end use, environmental impact, externalization
* DPSIR
* Clustering


The main objects are:
* Work session
* Case study
* Command

A work session can hold a registry of several command_executors.
A case study contains objects generated by command_executors. ¿References
A command can generate command_executors
A persisted case study must register sessions. It must register either a sequence of command_executors or a generator of command_executors (a Spreadsheet file, an R script). The three possibilities

"""


"""
The higher level object is a case study
It can be embedded in a file or registered
Each case study can have several IDs: OID, UUID, name
A sequence of command_executors
Modes of work
* exploratory elaboration of a version of a case study
* review of a version. READ ONLY command_executors, not added to the case study itself
* continue elaborating version of a case study (is a new version?)

Commands. A case study is evolved through command_executors which produce MuSIASEM primitives and provoke moving them, by ETL, by solving, output, etc.
Commands need to be adapted to the different interfaces, like Spreadsheet file or R scripts
Commands are executed in sequence, in transactions (the command_executors are grouped, then submitted in a pack)
Special command_executors to signal START of SUBMISSION and END of SUBMISSION of case study
Commands may be stored under a registered case study or simply executed and forgotten

Command
Several types. Compatibility with tool, ETL/specification/solving/visualization/export, process control (new case study, new submission, new transaction, login) or not
Metadata
Hierarchy

Pre-steps (in no particular order)
* Categories, taxonomies
* Mappings
* Dataset

Steps
* Identify compartments
* Identify relations
* Values
* Reasoning-Solving

"""

"""
* angular
* pyparsing - antlr
* python avanzado
* selenium

* estructuras, api, comandos, formato excel, persistencia

* estructura del proyecto en nexinfosys, incluyendo Python y R
* jupyterhub
* interacción (análisis), empaquetado para reproducibilidad y para demostración

* interface: restful. authentication (OAuth, JWS, OAuth2, other?)
* interface: navigation diagram
* excel: 
* command_executors: specification
* model: enumerate
* patterns: command, data structures, trace from command to generated entities, duck typing, jobs execution queue
* requirements: list of command_executors
* project structure. multiple languages, multiple modules, tests, README, setup, other?
* taiga 

"""

"""
almacenes (server URL)
casos de estudio (UUID)
envíos (ID, relative to the case study)
transacciones (ID, relative to the transaction)
comandos (ID, relative to the transaction)
primitivas (ID, relative to the command)

proc
fact
pf
valor. expresión

jerarq o lista
procs en jerarquía
fact en jerarquía


declare hierarchy1
add p
p is h1.asdf
p is h2.wer
p [p1 p2]
p3 [p1 p4]
(p1 € p y p3)

f [f1 f2]

add q

p.f1=3 m³*.F2
p.F2=5 m²

p1 > p2
p1.f1 > p2.f1

p1.f1 = (QQ)
p1.

swagger r
list command_executors, find an Excel expression

"""


"""
Persistent store, in database

* server (URL, UUID)
  * case study (UUID). The case study can be stored in several servers
    * submission (UUID). Several versions of the case study
      * transaction. A submission can be composed of several transactions
        * commands. A transaction is made of several commands
          * primitives. Commands can generate primitives. This should not be persisted, but generated in memory
          * command output. Command output may be stored
    * public primitives
  * users
  * groups
  * ACLs

"""
import datetime
import logging

from sqlalchemy import Column, Boolean, Integer, String, Unicode, DateTime, LargeBinary, ForeignKey
from copy import deepcopy
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm import relationship, backref, composite, scoped_session, sessionmaker, class_mapper, ColumnProperty, RelationshipProperty
from sqlalchemy.orm.state import InstanceState
from sqlalchemy.types import TypeDecorator, CHAR
# import psycopg2  # Only for automatic detection of dependencies
from sqlalchemy.dialects import postgresql
from sqlalchemy.dialects.postgresql import UUID, JSONB
import uuid
import jsonpickle

from nexinfosys.models import MODEL_VERSION
import sys

__author__ = 'rnebot'

DBSession = scoped_session(sessionmaker())  # TODO - Is this thread safe ??


class GUID(TypeDecorator):
    """
    Platform-independent GUID type.

    Uses Postgresql's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    """
    impl = CHAR

    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        # if dialect.name == 'postgresql':
        #     return uuid.UUID(value)
        # else:
        #     return value
        if value is None:
            return value
        else:
            return uuid.UUID(value)


def serialize_from_object(obj):
    tmp = sys.getrecursionlimit()
    sys.setrecursionlimit(10000)
    tmp_str = jsonpickle.encode(obj)  # .encode("ascii")
    sys.setrecursionlimit(tmp)
    return tmp_str


def deserialize_to_object(s):
    tmp = sys.getrecursionlimit()
    sys.setrecursionlimit(10000)
    tmp_str = jsonpickle.decode(s)
    sys.setrecursionlimit(tmp)
    return tmp_str


class BaseMixin(object):
    # query = DBSession.query_property()
    # @declared_attr
    # def __tablename__(cls):
    # return cls.__name__.lower()

    # def __new__(cls, *args, **kwargs):
    #     obj = super().__new__(cls)
    #     if '_sa_instance_state' not in obj.__dict__:
    #         obj._sa_instance_state = InstanceState(obj, obj._sa_class_manager)
    #     return obj

    def __deepcopy__(self, memo):
        cls = self.__class__
        logging.debug(str(cls))
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k in class_mapper(cls).iterate_properties:
            if isinstance(k, ColumnProperty):
                name = k.columns[0].name
                getattr(self, name, None)
            elif isinstance(k, RelationshipProperty):
                #if k.back_populates:
                name = k.strategy.key
                getattr(self, name, None)

        for k, v in self.__dict__.items():
            deepcopy(v, memo)
            # setattr(result, k, deepcopy(v, memo))
        return result

    # def __getstate__(self):
    #     state = self.__dict__.copy()
    #     if '_sa_instance_state':
    #         del state['_sa_instance_state']
    #     if isinstance(self, CaseStudy):
    #         if 'versions' in state:
    #             del state['versions']
    #     elif isinstance(self, CaseStudyVersion):
    #         if 'sessions' in state:
    #             del state['sessions']
    #     elif isinstance(self, CaseStudyVersionSession):
    #         if 'commands' in state:
    #             del state['commands']
    #     elif isinstance(self, CommandsContainer):
    #         if 'children_commands' in state:
    #             del state['children_commands']
    #     elif isinstance(self, User):
    #         if 'authenticators' in state:
    #             del state['authenticators']
    #         if 'groups' in state:
    #             del state['groups']
    #     return state
    #
    # def __setstate__(self, state):
    #     self.__dict__.update(state)

ORMBase = declarative_base(cls=BaseMixin)

# dictalchemy.make_class_dictable(ORMBase)

# ---------------------------------------------------------------------------------------------------------------------


def force_load(obj1):

    def force_load_int(obj):
        cls = obj.__class__
        result = cls.__new__(cls)
        memo[id(obj)] = result
        for k in class_mapper(cls).iterate_properties:
            if isinstance(k, ColumnProperty):
                name = k.columns[0].name
                getattr(obj, name, None)
            elif isinstance(k, RelationshipProperty):
                name = k.strategy.key
                v = getattr(obj, name, None)
                if v:
                    try:
                        for it in v:
                            if id(it) not in memo:
                                force_load_int(it)
                    except TypeError:
                        if id(v) not in memo:
                            force_load_int(v)
    memo = {}
    force_load_int(obj1)
    return obj1

# ---------------------------------------------------------------------------------------------------------------------
# Authentication
# Authorization
#


class Authenticator(ORMBase):  # CODES
    """ List of valid authenticators """
    __tablename__ = "authenticators"

    id = Column(Integer, primary_key=True, autoincrement=True)
    uuid = Column(GUID, unique=True)
    name = Column(String(80))
    validation_endpoint = Column(String(1024))


class User(ORMBase):
    """ Identities.
     Users can have one or more authenticators
     Permissions (ACLs) are assigned to Users or Groups of users """
    __tablename__ = "users"

    id = Column(Integer, primary_key=True, autoincrement=True)
    uuid = Column(GUID, unique=True)
    name = Column(String(80))
    email = Column(String(80))
    creation_time = Column(DateTime, default=datetime.datetime.utcnow())
    deactivation_time = Column(DateTime)


class UserAuthenticator(ORMBase):
    """ Recognized user authenticator """
    __tablename__ = "users_authenticators"

    user_id = Column(Integer, ForeignKey(User.id), nullable=False, primary_key=True)
    user = relationship(User, backref=backref("authenticators", cascade="all, delete-orphan"))
    authenticator_id = Column(Integer, ForeignKey(Authenticator.id), nullable=False, primary_key=True)
    authenticator = relationship(Authenticator)

    email = Column(String(80))


class Group(ORMBase):
    __tablename__ = "groups"

    id = Column(Integer, primary_key=True, autoincrement=True)
    uuid = Column(GUID, unique=True)
    name = Column(String(80))


class GroupUser(ORMBase):
    __tablename__ = "groups_users"
    group_id = Column(Integer, ForeignKey(Group.id), nullable=False, primary_key=True)
    user_id = Column(Integer, ForeignKey(User.id), nullable=False, primary_key=True)
    group = relationship(Group, backref=backref("users", cascade="all, delete-orphan"))
    user = relationship(User, backref=backref("groups", cascade="all, delete-orphan"))


class ObjectType(ORMBase):  # CODES
    """ Processor, Flow or Fund (Factor), Hierarchy, ... """
    __tablename__ = "object_types"

    id = Column(Integer, primary_key=True, autoincrement=True)
    uuid = Column(GUID, unique=True)
    name = Column(String(80), nullable=False)


class PermissionType(ORMBase):  # CODES
    __tablename__ = "permission_types"

    id = Column(Integer, primary_key=True, autoincrement=True)
    uuid = Column(GUID, nullable=False)  # Object ID (ACL are on objects with UUID)
    name = Column(String(80))


class ACL(ORMBase):
    """ List of permissions on an object. The detail """
    __tablename__ = "permissions"

    id = Column(Integer, primary_key=True, autoincrement=True)
    uuid = Column(GUID, nullable=False)  # Object ID (ACL are on objects with UUID)
    object_type = Column(Integer, ForeignKey(ObjectType.id), nullable=False)
    object_id = Column(GUID, nullable=False)


class ACLDetail(ORMBase):
    __tablename__ = "permissions_detail"

    id = Column(Integer, primary_key=True, autoincrement=True)
    acl_id = Column(Integer, ForeignKey(ACL.id))
    acl = relationship(ACL, backref=backref("detail", cascade="all, delete-orphan"))
    user = Column(Integer, ForeignKey(User.id))
    group = Column(Integer, ForeignKey(Group.id))
    permission_id = Column(Integer, ForeignKey(PermissionType.id))  # Read, Export (Share?), Modify, Delete (depends on the type of object)
    permission = relationship(PermissionType)

# ---------------------------------------------------------------------------------------------------------------------


class Server(ORMBase):
    """ In case there are federated servers (for the future). The use of UUIDs is related to this. """
    __tablename__ = "servers"

    id = Column(Integer, primary_key=True, autoincrement=True)
    uuid = Column(GUID, unique=True)
    url = Column(String(1024))  # RESTful endpoint of remote server
    cached_name = Column(String(80))
    cached_description = Column(String(1024))

# ---------------------------------------------------------------------------------------------------------------------


class PedigreeTemplate(ORMBase):
    __tablename__ = "nusap_pedigree_templates"

    id = Column(Integer, primary_key=True, autoincrement=True)
    uuid = Column(GUID, unique=True)
    # TODO A list of lists. Each master list has a position, from 1 to n, n being the number of lists, without jumps
    # TODO These lists have a name and a description
    # TODO Each element of a contained list has also a position, plus a short label and a description

# ---------------------------------------------------------------------------------------------------------------------
# Objects supporting "process" in case studies (evolution in time, improvement, collaboration, tracking of changes,...)
# ---------------------------------------------------------------------------------------------------------------------


class CaseStudyStatus(ORMBase):  # CODES
    __tablename__ = "cs_cod_status_codes"

    id = Column(Integer, primary_key=True, autoincrement=True)
    uuid = Column(GUID, unique=True)
    # ELABORATION, FINISHED, FINISHED_PUBLISHED
    name = Column(String(80), nullable=False)

"""
Two special case studies
* One to contain shared objects, like grammars and hierarchies
* Other for unregistered tests (those with no metadata). Case studies 
"""


class CaseStudy(ORMBase):
    __tablename__ = "cs"

    id = Column(Integer, primary_key=True, autoincrement=True)
    uuid = Column(GUID, unique=True, default=str(uuid.uuid4()))
    name = Column(String(80), default=None)
    oid = Column(String(128), nullable=True)
    internal_code = Column(String(20), nullable=True)
    description = Column(String(1024), default=None)
    areas = Column(String(10), default=None)  # W(ater), E(nergy), F(ood), L(and), C(limate -change-), ...
    geographic_level = Column(String(4), nullable=True)  # R(egional), C(ountry), E(urope), S(ectoral)
    restriction_level = Column(String(3), nullable=True)  # I(nternal), C(onfidential), P(ublic)
    version = Column(String(10), nullable=True)

    def __copy__(self):
        cs = CaseStudy()
        # CODES are NOT copied !!!
        cs.uuid = str(uuid.uuid4())  # NEW UUID !!!!
        cs.oid = None
        cs.internal_code = None
        # Other fields (not CODES)
        cs.name = self.name
        cs.description = self.description
        cs.areas = self.areas
        cs.restriction_level = self.restriction_level
        cs.geographic_level = self.geographic_level
        cs.version = self.version

        return cs

    def __init__(self):
        if not self.uuid:
            self.uuid = str(uuid.uuid4())

# class Diagram(ORMBase):
#     __tablename__ = 'cs_diagrams'
#
#     id = Column(Integer, primary_key=True, autoincrement=True)
#     page = Column(String(80), nullable=False)
#     description = Column(String(512), nullable=True)
#     content = Column(JSON, nullable=True)


class CaseStudyVersion(ORMBase):
    """
    A case study can evolve. Different versions
    """
    __tablename__ = "cs_versions"

    id = Column(Integer, primary_key=True, autoincrement=True)
    uuid = Column(GUID, unique=True, default=str(uuid.uuid4()))
    name = Column(String(256), default=None)

    case_study_id = Column(Integer, ForeignKey(CaseStudy.id))
    case_study = relationship(CaseStudy, backref=backref("versions", cascade="all, delete-orphan"))

    creation_instant = Column(DateTime, default=None)

    status_id = Column(Integer, ForeignKey(CaseStudyStatus.id))
    status = relationship(CaseStudyStatus)

    issues = Column(Unicode, nullable=True, default=None)  # If the session resulted in issues (warnings, errors,...) register here
    state = Column(Unicode, nullable=True, default=None)  # Persisted state after the execution of all commands

    state_version = Column(Integer, nullable=True, default=MODEL_VERSION)

    def __copy__(self):
        vs = CaseStudyVersion()
        # vs.id = # NOT copied, a new ID !!!
        # vs.uuid = # NOT copied, a new UUID !!!!
        vs.name = self.name
        # vs.case_study =  NOT COPIED, it must be assigned after COPY !!!!
        # vs.creation_instant = NOT COPIED, the default value -current time- is kept !!!!
        vs.status_id = self.status_id
        return vs

    def __init__(self):
        # if "_sa_instance_state" not in self.__dict__:
        #     self.__dict__.update({"_sa_instance_state": InstanceState(self, self._sa_class_manager)})
        if not self.uuid:
            self.uuid = str(uuid.uuid4())


class CaseStudyVersionSession(ORMBase):
    """
    A version is elaborated in different steps, which are called "work sessions"
    
    """
    __tablename__ = "cs_versions_sessions"

    id = Column(Integer, primary_key=True, autoincrement=True)
    uuid = Column(GUID, unique=True, default=str(uuid.uuid4()))

    version_id = Column(Integer, ForeignKey(CaseStudyVersion.id))
    version = relationship(CaseStudyVersion, backref=backref("sessions", order_by="CaseStudyVersionSession.open_instant", cascade="all, delete-orphan"))

    # If FALSE, this session loads and resumes state from previous sessions. TRUE, start from "zero" status
    restarts = Column(Boolean, default=False)  # Default is "continue" with previous state

    who_id = Column(Integer, ForeignKey(User.id), nullable=False)
    who = relationship(User)

    open_instant = Column(DateTime, default=datetime.datetime.utcnow())
    close_instant = Column(DateTime, default=None)

    # Outcomes of the execution of commands in the Session
    issues = Column(Unicode, nullable=True, default=None)  # If the session resulted in issues (warnings, errors,...) register here
    state = Column(Unicode, nullable=True, default=None)  # Persisted state after the execution of all commands

    state_version = Column(Integer, nullable=True, default=MODEL_VERSION)

    def __init__(self):
        if not self.uuid:
            self.uuid = str(uuid.uuid4())

    def __copy__(self):
        s = CaseStudyVersionSession()
        # s.uuid = NOT copied, a new UUID !!!!
        # s.version = NOT copied, it has to be assigned !!!
        s.restarts = False  # Always FALSE !!!
        s.who = self.who
        s.open_instant = self.open_instant
        s.close_instant = self.close_instant
        s.issues = self.issues
        s.state = self.state
        s.state_version = self.state_version
        return s


class CommandsContainer(ORMBase):
    """ The CommandsContainer object is either the representation of a primitive command or of a generator of command_executors.
        It is not the executable command
        For a primitive command, the content is the serialized JSON content elaborated by the command itself.
        In this case, generator_type is "primitive", content_type "application/json"
        For command generators, the content may be a Spreadsheet file or an R script.
        The generator_type could also be "Spreadsheet", "Rscript".
        The content type for the first would be "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet".
    """
    __tablename__ = "cs_versions_sessions_cmd_gens"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(80))

    session_id = Column(Integer, ForeignKey(CaseStudyVersionSession.id))
    session = relationship(CaseStudyVersionSession, backref=backref("commands", order_by="CommandsContainer.order",
                           collection_class=ordering_list('order'), cascade="all, delete-orphan"))

    order = Column(Integer)
    generator_type = Column(String(48))  # "Primitive", "Spreadsheet", "Rscript"
    content_type = Column(String(128))  # "application/json", "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet", "application/text"
    content = Column(LargeBinary, nullable=True)  # A JSON string, a Worksheet file, an R script, ...

    # When the command was executed (the result is in the state of the version)
    execution_start = Column(DateTime, nullable=True)
    execution_end = Column(DateTime, nullable=True)

    parent_command_id = Column(Integer, ForeignKey("cs_versions_sessions_cmd_gens.id"))
    parent_command = relationship("CommandsContainer", backref=backref("children_commands", remote_side=[id], order_by="CommandsContainer.order",
                                  collection_class=ordering_list('order')), cascade="all, delete-orphan")
    # Role of the command regarding its parent command (to make child commands independent of the order,
    # like "named parameters", or "iterated command")
    parent_role = Column(String(128))

    @staticmethod
    def create(generator_type, file_type, file):
        """
        Generates command_executors from an input stream (string or file)
        There must be a factory to parse stream 
        :param generator_type: 
        :param file_type: 
        :param file: It can be a stream or a URL or a file name
        """
        c = CommandsContainer()
        c.name = None
        c.generator_type = generator_type
        c.content_type = file_type
        c.content = file  # TODO Read all the content IN-MEMORY
        # c.parent_command # TODO Commands do not allow parents for now
        c.parent_role = None
        c.execution_start = None
        c.execution_end = None
        return c

    def __copy__(self):
        c = CommandsContainer()
        c.name = self.name
        # c.session NOT COPIED, assign it directly !!!!
        c.order = self.order
        c.generator_type = self.generator_type
        c.content_type = self.content_type
        c.content = self.content
        c.parent_command_id = None
        # TODO c.parent_command_id = NOT COPIED, assign it directly !!!!
        # TODO or Copy children
        c.parent_role = self.parent_role
        return c


def load_table(sf, clazz, d):
    """
    Insert pairs (key, value) into a relational table
    It loads a dictionary "d" containing keys and values, into a relational table associated to the class "clazz",
    using the session factory "sf"
    
    :param sf: 
    :param clazz: 
    :param d: 
    :return: 
    """
    session = sf()
    for k, v in d.items():
        i = session.query(clazz).filter(clazz.uuid == k).first()
        if not i:
            ins = clazz()
            ins.uuid = k
            ins.name = v
            session.add(ins)
    session.commit()
    sf.remove()



import logging

from flask import Flask

from nexinfosys import initialize_configuration, cfg_file_env_var
from nexinfosys.model_services import get_case_study_registry_objects

nis_api_base = "/nis_api"  # Base for all RESTful calls
nis_client_base = "/nis_client"  # Base for the Angular2 client
nis_external_client_base = "/nis_external"  # Base for the Angular2 client called from outside

app = Flask(__name__)
app.debug = True
UPLOAD_FOLDER = '/tmp/'
app.config['UPLOAD_FOLDER'] = UPLOAD_FOLDER

# Initialize configuration
initialize_configuration()
app.config.from_envvar(cfg_file_env_var)


def get_results_in_session(isess: "InteractiveSession"):
    """
    Obtain list of ALL possible outputs (not only datasets) IN the current state.
    :param isess:
    :return:
    """
    dataset_formats = ["CSV", "XLSX", "SDMX.json"] #, "XLSXwithPivotTable", "NISembedded", "NISdetached"]
    graph_formats = ["VisJS", "GML"] #, "GraphML"]
    ontology_formats = ["OWL"]
    geo_formats = ["GeoJSON"]
    # A reproducible session must be open, signal about it if not
    if isess.reproducible_session_opened():
        if isess.state:
            glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(isess.state)
            r = {"datasets":
                     [dict(name=k,
                           type="dataset",
                           description=F"{datasets[k].description} [{datasets[k].data.shape[0]} rows, {datasets[k].data.size} cells, "
                           F"{datasets[k].data.memory_usage(True).sum()} bytes]",
                           # nelements=datasets[k].data.size,
                           # nrows=datasets[k].data.shape[0],
                           # size=datasets[k].data.memory_usage(True).sum(),
                           formats=[dict(format=f,
                                         url=nis_api_base + F"/isession/rsession/state_query/datasets/{k}.{f.lower()}")
                                    for f in dataset_formats],
                           ) for k in datasets
                      ] +
                     [dict(name="interfaces_graph",
                           type="graph",
                           description="Graph of Interfaces, Quantities; Scales and Exchanges",
                           formats=[dict(format=f,
                                         url=nis_api_base + F"/isession/rsession/state_query/flow_graph.{f.lower()}")
                                    for f in graph_formats]),
                      dict(name="processors_graph",
                           type="graph",
                           description="Processors and exchanges graph",
                           formats=[dict(format=f,
                                         url=nis_api_base + F"/isession/rsession/state_query/processors_graph.{f.lower()}")
                                    for f in graph_formats]),
                      ] +
                     [dict(name="Sankey_Graph",
                           type="Graph",
                           description="Dictionary of Sankey Graph for every scenario for implementation in JupyterLab using plotly",
                           formats=[
                               dict(format=f, url=nis_api_base + F"/isession/rsession/state_query/sankey_graph.{f.lower()}")
                               for f in ["JSON"]]),
                      ] +
                     [dict(name="processors_geolayer",
                           type="geolayer",
                           description="Processors",
                           formats=[
                               dict(format=f, url=nis_api_base + F"/isession/rsession/state_query/geolayer.{f.lower()}")
                               for f in geo_formats]),
                      ] +
                     [dict(name="Model",
                           type="model",
                           description="Model",
                           formats=[
                               dict(format=f, url=nis_api_base + F"/isession/rsession/state_query/model.{f.lower()}")
                               for f in ["JSON", "XLSX", "XML"]]),
                      ] +
                     [dict(name="Ontology",
                           type="ontology",
                           description="OWL ontology",
                           formats=[
                               dict(format=f, url=nis_api_base + F"/isession/rsession/state_query/ontology.{f.lower()}")
                               for f in ontology_formats]),
                      ] +
                     [dict(name="Python script",
                           type="script",
                           description="Python script",
                           formats=[dict(format=f,
                                         url=nis_api_base + F"/isession/rsession/state_query/python_script.{f.lower()}")
                                    for f in ["Python", "JupyterNotebook"]]),
                      dict(name="R script",
                           type="script",
                           description="R script",
                           formats=[
                               dict(format=f, url=nis_api_base + F"/isession/rsession/state_query/r_script.{f.lower()}")
                               for f in ["R", "JupyterNotebook"]]),
                      ] +
                     [dict(name="Commands reference",
                           type="document",
                           description="Reference of all commands and their fields",
                           formats=[
                               dict(format=f,
                                    url=nis_api_base + F"/isession/rsession/state_query/commands_reference_document.{f}")
                               for f in ["html"]]),
                      ]
                 }

    return r


from enum import Enum
from attr import attrs, attrib, validators
from openpyxl.utils import get_column_letter

import nexinfosys
from nexinfosys.common.helper import create_dictionary


class IType(Enum):
    INFO = 1
    WARNING = 2
    ERROR = 3


@attrs
class IssueLocation:
    sheet_name = attrib()
    sheet_number = attrib(default=None)
    row = attrib(default=None)
    column = attrib(default=None)

    def __str__(self):
        return f'(sheet_name="{self.sheet_name}", sheet_no={self.sheet_number}, row={self.row}, ' \
               f'column={get_column_letter(self.column) if self.column else "-"})'


@attrs
class Issue:
    # (1) Info, (2) Warning, (3) Error
    itype = attrib()  # type: IType
    # An english description of what happened
    description = attrib()  # type: str
    # Command type
    ctype = attrib(default=None)  # type: str
    # Where is the issue. The expression of the location depends. For spreadsheet it is sheet name, row and column
    location = attrib(default=None)  # type: IssueLocation

    def __str__(self):
        return f'(level={self.itype.name}, msg="{self.description}", cmd="{self.ctype}", {self.location})'

    @itype.validator
    def check(self, attribute, value):
        if not isinstance(value, nexinfosys.command_generators.IType):
            raise ValueError("itype should be an instance of enum IType")


# List of Global Functions (used by "parser_field_parsers.py" and "parser_ast_evaluators.py")
global_functions = create_dictionary(data={i["name"]: i for i in
                    [{"name": "cos",
                      "full_name": "math.cos",
                      "kwargs": {}},
                     {"name": "sin",
                      "full_name": "math.sin",
                      "kwargs": {}},
                     {"name": "InterfaceType",
                      "full_name": "nexinfosys.command_generators.parser_ast_evaluators.get_interface_type",
                      "kwargs": {},
                      "special_kwargs": {"PartialRetrievalDictionary": "prd"}},
                     {"name": "NISName",
                      "full_name": "nexinfosys.command_generators.parser_ast_evaluators.get_nis_name",
                      "kwargs": {}},
                     {"name": "Processor",
                      "full_name": "nexinfosys.command_generators.parser_ast_evaluators.get_processor",
                      "kwargs": {},
                      "special_kwargs": {"PartialRetrievalDictionary": "prd"}},
                     {"name": "LCIAMethod",
                      "full_name": "nexinfosys.command_generators.parser_ast_evaluators.lcia_method",
                      "kwargs": {},
                      "special_kwargs": {"IndicatorState": "state",
                                         "LCIAMethods": "lcia_methods"}},
                     {"name": "UDIF",  # "User Defined Indicator Function"
                      "full_name": "nexinfosys.command_generators.parser_ast_evaluators.call_udif_function",
                      "kwargs": {},
                      "special_kwargs": {"IndicatorState": "state"}},
                     ]
                    })

global_functions_extended = create_dictionary(data={i["name"]: i for i in
                    [{"name": "sum",
                      "full_name": "nexinfosys.command_generators.parser_ast_evaluators.aggregator_sum",
                      "kwargs": {},
                      "special_kwargs": {"ProcessorsMap": "processors_map", "ProcessorsDOM": "processors_dom",
                                         "DataFrameGroup": "df_group", "IndicatorsDataFrameGroup": "df_indicators_group"}},
                     {"name": "avg",
                      "full_name": "nexinfosys.command_generators.parser_ast_evaluators.aggregator_avg",
                      "kwargs": {},
                      "special_kwargs": {"ProcessorsMap": "processors_map", "ProcessorsDOM": "processors_dom",
                                         "DataFrameGroup": "df_group", "IndicatorsDataFrameGroup": "df_indicators_group"}},
                     {"name": "max",
                      "full_name": "nexinfosys.command_generators.parser_ast_evaluators.aggregator_max",
                      "kwargs": {},
                      "special_kwargs": {"ProcessorsMap": "processors_map", "ProcessorsDOM": "processors_dom",
                                         "DataFrameGroup": "df_group", "IndicatorsDataFrameGroup": "df_indicators_group"}},
                     {"name": "min",
                      "full_name": "nexinfosys.command_generators.parser_ast_evaluators.aggregator_min",
                      "kwargs": {},
                      "special_kwargs": {"ProcessorsMap": "processors_map", "ProcessorsDOM": "processors_dom",
                                         "DataFrameGroup": "df_group", "IndicatorsDataFrameGroup": "df_indicators_group"}},
                     {"name": "count",
                      "full_name": "nexinfosys.command_generators.parser_ast_evaluators.aggregator_count",
                      "kwargs": {},
                      "special_kwargs": {"ProcessorsMap": "processors_map", "ProcessorsDOM": "processors_dom",
                                         "DataFrameGroup": "df_group", "IndicatorsDataFrameGroup": "df_indicators_group"}},
                     {"name": "nancount",
                      "full_name": "nexinfosys.command_generators.parser_ast_evaluators.aggregator_nan_count",
                      "kwargs": {},
                      "special_kwargs": {"ProcessorsMap": "processors_map", "ProcessorsDOM": "processors_dom",
                                         "DataFrameGroup": "df_group", "IndicatorsDataFrameGroup": "df_indicators_group"}},
                     ]
                    })

_ = [k for k in global_functions.keys()]+[k for k in global_functions_extended.keys()]
extended_dict_of_function_names = create_dictionary(data={k: None for k in _})
_ = extended_dict_of_function_names


from enum import Enum
from typing import TypeVar

from nexinfosys.common.helper import FloatExp

Node = TypeVar('Node')  # Generic node type
Weight = FloatExp  # Type alias
Value = FloatExp  # Type alias


class EdgeType(Enum):
    """ Type of edge of a ComputationGraph """
    DIRECT = 0
    REVERSE = 1



from collections import namedtuple

Domain = namedtuple("Domain", "type values")
Field = namedtuple("Field", "name domain mandatory")

ref_prof = [{"type": "geographic",  # Reduced, from ISO19139
             "fields":
                 [Field(name="title", domain=Domain(type="string", values=None), mandatory="True"),
                  Field(name="date", domain=Domain(type="string", values=None), mandatory="True"),
                  Field(name="boundingbox", domain=Domain(type="string", values=None), mandatory="False"),
                  # Field(name="language", domain=Domain(type="string", values=None), mandatory="True"),
                  # Field(name="charset", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="topiccategory", domain=Domain(type="string", values=None), mandatory="True"),
                  # Field(name="metadatalanguage", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="description", domain=Domain(type="string", values=None), mandatory="True"),
                  # Field(name="metadatacharset", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="metadatapointofcontact", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="annote", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="datalocation", domain=Domain(type="string", values=None), mandatory="True"),
                  # This is not ISO19139. Reference to definition of data
                  ]
             },
            {"type": "provenance",  # Reduced, from W3C Provenance Recommendation
             "fields":
                 [Field(name="agenttype", domain=Domain(type="string", values=["person", "software", "organization"]),
                        mandatory="True"),
                  Field(name="agent", domain=Domain(type="string", values=None), mandatory="True"),
                  Field(name="activities", domain=Domain(type="string", values=None), mandatory="True"),
                  Field(name="entities", domain=Domain(type="string", values=None), mandatory="False"),
                  ]
             },
            {"type": "bibliographic",  # From BibTex
             "fields":
                 [Field(name="entry_type", domain=Domain(type="string",
                                                         values=["article", "book", "booklet", "conference", "inbook",
                                                                 "incollection", "inproceedings", "mastersthesis", "misc",
                                                                 "phdtesis", "proceedings", "techreport", "unpublished"]),
                        mandatory="True"),
                  Field(name="address", domain=Domain(type="string", values=None),
                        mandatory="entry_type not in ('article', 'misc')"),
                  Field(name="annote", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="author", domain=Domain(type="string", values=None), mandatory="entry_type not in ('misc')"),
                  Field(name="booktitle", domain=Domain(type="string", values=None),
                        mandatory="entry_type in ('incollection', 'inproceedings')"),
                  Field(name="chapter", domain=Domain(type="string", values=None),
                        mandatory="entry_type in ('inbook', 'incollection')"),
                  Field(name="crossref", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="edition", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="editor", domain=Domain(type="string", values=None),
                        mandatory="entry_type in ('book', 'inbook', 'incollection', 'inproceedings', 'proceedings')"),
                  Field(name="howpublished", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="institution", domain=Domain(type="string", values=None),
                        mandatory="entry_type in ('techreport')"),
                  Field(name="journal", domain=Domain(type="string", values=None), mandatory="entry_type in ('article')"),
                  Field(name="key", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="month", domain=Domain(type="string",
                                                    values=["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep",
                                                            "Oct", "Nov", "Dec"]), mandatory="False"),
                  Field(name="note", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="number", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="organization", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="pages", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="publisher", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="school", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="series", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="title", domain=Domain(type="string", values=None), mandatory="entry_type not in ('misc')"),
                  Field(name="type", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="url", domain=Domain(type="string", values=None), mandatory="False"),
                  # This one is not BibTex standard, but it is necessary
                  Field(name="volume", domain=Domain(type="string", values=None), mandatory="False"),
                  Field(name="year", domain=Domain(type="int", values=None),
                        mandatory="entry_type in ('article', 'book', 'inbook', 'incollection', 'inproceedings', 'mastersthesis', 'phdthesis', 'proceedings', 'techreport')")
                  ]
             }
            ]


def profile_field_name_sets():
    m = {}
    for prof in ref_prof:
        m[prof["type"]] = set([i.name.lower() for i in prof["fields"]])
    return m



import json
import math

from nexinfosys.common.helper import strcmp
from nexinfosys.model_services import IExecutableCommand, State, get_case_study_registry_objects
from nexinfosys.models.musiasem_concepts import Observer, Processor, \
    FactorsRelationScaleObservation, \
    ProcessorsRelationPartOfObservation, \
    ProcessorsRelationUpscaleObservation, \
    ProcessorsSet


class UpscaleCommand(IExecutableCommand):
    """
    Serves to instantiate processor templates inside another processor
    See Almeria case study for the base example of this
    """
    def __init__(self, name: str):
        self._name = name
        self._content = None

    def execute(self, state: "State"):
        """
        For each parent processor clone all the child processors.
        The cloning process may pass some factor observation, that may result in
        """
        some_error = False
        issues = []

        parent_processor_type = self._content["parent_processor_type"]
        child_processor_type = self._content["child_processor_type"]
        scaled_factor = self._content["scaled_factor"]
        source = self._content["source"]
        # column_headers = self._content["column_headers"]
        # row_headers = self._content["row_headers"]
        scales = self._content["scales"]

        # Find processor sets, for parent and child
        glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)
        if parent_processor_type not in p_sets:
            some_error = True
            issues.append((3, "The processor type '"+parent_processor_type +
                           "' (appointed for parent) has not been found in the commands execute so far"))

        if child_processor_type not in p_sets:
            some_error = True
            issues.append((3, "The processor type '"+child_processor_type +
                           "' (should be child processor) has not been found in the commands execute so far"))

        if some_error:
            return issues, None

        # CREATE the Observer of the Upscaling
        oer = glb_idx.get(Observer.partial_key(source))
        if not oer:
            oer = Observer(source)
            glb_idx.put(oer.key(), oer)
        else:
            oer = oer[0]

        # Processor Sets have associated attributes, and each of them has a code list
        parent = p_sets[parent_processor_type]  # type: ProcessorsSet
        child = p_sets[child_processor_type]  # type: ProcessorsSet

        # Form code lists from the command specification
        code_lists = None
        for sc_dict in scales:
            codes = sc_dict["codes"]
            if not code_lists:
                code_lists = [set() for _ in codes]

            for i, c in enumerate(codes):
                code_lists[i].add(c)

        # Match existing code lists (from Processor attributes) with the ones gathered in the specification of
        # the two (parent and child) processors sets.
        # Form lists of attributes of processors used in the code lists
        parent_attrs = []
        child_attrs = []
        matched = []
        for i, cl in enumerate(code_lists):
            found = False
            for attr, attr_values in parent.attributes.items():
                if set(attr_values).issuperset(cl):
                    parent_attrs.append((attr, i))  # (Attribute, code list index)
                    found = True
                    break
            for attr, attr_values in child.attributes.items():
                if set(attr_values).issuperset(cl):
                    child_attrs.append((attr, i))  # (Attribute, code list index)
                    found = True
                    break
            matched.append(found)
        for i, found in enumerate(matched):
            if not found:
                cl = code_lists[i]
                # TODO Try cl as a list of names of parent or child processors
                if not found:
                    issues.append((2, "The code list: " + ", ".join(cl) + " is not contained in the attributes of the parent processors set '" + parent_processor_type + "' nor in the attributes of the child processors set '" + child_processor_type + "'"))

        # Execute the upscale for each
        cached_processors = {}
        for sc_dict in scales:
            try:
                non_zero_weight = math.fabs(float(sc_dict["weight"])) > 1e-6
            except:
                non_zero_weight = True
            if not non_zero_weight:
                continue

            codes = sc_dict["codes"]
            # Find parent processor
            parent_dict = {attr: codes[i] for attr, i in parent_attrs}
            d2s = str(parent_dict)
            if d2s in cached_processors:
                parent = cached_processors[d2s]
                if not parent:
                    issues.append((3, "Either the tuple (" + d2s + ") did not match any Processor or matched more than one."))
            else:
                parent_dict.update(Processor.partial_key())

                # Obtain Processor matching the attributes <<<<<<<<<<
                # Query the PartialRetrievalDictionary by attributes
                parents = glb_idx.get(parent_dict)

                if len(parents) > 1:
                    issues.append((3, "The tuple ("+str(parent_dict)+") matches "+str(len(parents))+" Processors: "+(", ".join([p.name for p in parents]))))
                    parent = None
                elif len(parents) == 0:
                    issues.append((3, "The tuple (" + str(parent_dict) + ") did not match any Processor"))
                    parent = None
                else:
                    parent = parents[0]

                cached_processors[d2s] = parent

            # Find child processor
            child_dict = {attr: codes[i] for attr, i in child_attrs}
            d2s = str(child_dict)
            if d2s in cached_processors:
                child = cached_processors[d2s]
                if not child:
                    issues.append((3, "Either the tuple (" + d2s + ") did not match any Processor or matched more than one."))
            else:
                child_dict.update(Processor.partial_key())

                # Obtain Processors matching the attributes
                # Query the PartialRetrievalDictionary by attributes
                children = glb_idx.get(child_dict)

                if len(children) > 1:
                    issues.append((3, "The tuple ("+str(child_dict)+") matches "+str(len(parents))+" Processors: "+(", ".join([p.name for p in children]))))
                    child = None
                elif len(children) == 0:
                    issues.append((3, "The tuple (" + str(child_dict) + ") did not match any Processor"))
                    child = None
                else:
                    child = children[0]  # type: Processor

                cached_processors[d2s] = child

            # Clone child processor (and its descendants) and add an upscale relation between "parent" and the clone
            if parent and child:
                if non_zero_weight:
                    # Clone the child processor
                    # TODO
                    cloned_child, cloned_children = child.clone(state=glb_idx)
                    Processor.register([cloned_child] + list(cloned_children), glb_idx)

                    # Create the new Relation Observations
                    # - Part-of Relation
                    o1 = ProcessorsRelationPartOfObservation.create_and_append(parent, cloned_child, oer)  # Part-of
                    glb_idx.put(o1.key(), o1)
                    # - Upscale Relation
                    quantity = str(sc_dict["weight"])
                    if True:
                        # Find Interface named "scaled_factor"
                        for f in parent.factors:
                            if strcmp(f.name, scaled_factor):
                                origin = f
                                break
                        else:
                            origin = None
                        for f in cloned_child.factors:
                            if strcmp(f.name, scaled_factor):
                                destination = f
                                break
                        else:
                            destination = None

                        if origin and destination:
                            o3 = FactorsRelationScaleObservation.create_and_append(origin, destination,
                                                                                   observer=None,
                                                                                   quantity=quantity)
                            glb_idx.put(o3.key(), o3)
                        else:
                            raise Exception("Could not find Interfaces to define a Scale relation. Processors: " +
                                            parent.name+", "+cloned_child.name+"; Interface name: "+scaled_factor)
                    else:
                        o3 = ProcessorsRelationUpscaleObservation.create_and_append(parent, cloned_child,
                                                                                    observer=None,
                                                                                    factor_name=scaled_factor,
                                                                                    quantity=quantity)
                        glb_idx.put(o3.key(), o3)
            else:
                # TODO
                parent_dict = str({attr: codes[i] for attr, i in parent_attrs})
                child_dict = str({attr: codes[i] for attr, i in child_attrs})
                if not parent and child:
                    issues.append((2, "Could not find parent Processor matching attributes: "+parent_dict))
                elif not child and parent:
                    issues.append((2, "Could not find child Processor matching attributes: "+child_dict))
                else:
                    issues.append((2, "Could not find parent Processor matching attributes: "+parent_dict+", nor child Processor matching attributes: " + child_dict))

        return issues, None

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the content
        return self._content

    def json_deserialize(self, json_input):
        # TODO Check validity
        issues = []
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)
        return issues



import json

from nexinfosys.command_generators import Issue, IssueLocation, IType
from nexinfosys.model_services import IExecutableCommand, get_case_study_registry_objects
from nexinfosys.models.musiasem_concepts import ProvenanceReference, BibliographicReference, \
    GeographicReference, Observer


class ReferencesV2Command(IExecutableCommand):
    """ It is a mere data container
        Depending on the type, the format can be controlled with a predefined schema
    """
    def __init__(self, name: str, ref_type: type):
        self._name = name
        self._content = None
        self._ref_type = ref_type

    @property
    def ref_type(self):
        return self._ref_type

    def execute(self, state: "State"):
        """
        Process each of the references, simply storing them as Reference objects
        """
        glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)
        name = self._content["command_name"]
        issues = []

        # Receive a list of validated references
        # Store them as objects, which can be referred to later
        for ref in self._content["items"]:
            r = ref["_row"]

            if "ref_id" not in ref:
                issues.append(Issue(itype=IType.ERROR,
                                    description="'ref_id' field not found: "+str(ref),
                                    location=IssueLocation(sheet_name=name, row=r, column=None)))
                continue
            else:
                ref_id = ref["ref_id"]
                existing = glb_idx.get(self.ref_type.partial_key(ref_id))
                if len(existing) == 1:
                    issues.append(Issue(itype=IType.ERROR,
                                        description="Reference '"+ref_id+"' of type '"+str(self.ref_type)+"' is already defined. Not allowed",
                                        location=IssueLocation(sheet_name=name, row=r, column=None)))
                    continue
                elif len(existing) > 1:  # This condition should not occur...
                    issues.append(Issue(itype=IType.ERROR,
                                        description="The reference '"+ref_id+"' of type '"+str(self.ref_type)+"' is defined more than one time ("+str(len(existing))+")",
                                        location=IssueLocation(sheet_name=name, row=r, column=None)))
                    continue

                # Create and store the Reference
                reference = self.ref_type(ref_id, ref)
                glb_idx.put(reference.key(), reference)

                # BibliographicReference and ProvenanceReference ar also Observer
                if isinstance(reference, Observer):
                    glb_idx.put(Observer.key(reference), reference)

        return issues, None

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the content
        return self._content

    def json_deserialize(self, json_input):
        # TODO Check validity
        issues = []
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)
        return issues


class ProvenanceReferencesCommand(ReferencesV2Command):
    def __init__(self, name: str):
        ReferencesV2Command.__init__(self, name, ProvenanceReference)


class BibliographicReferencesCommand(ReferencesV2Command):
    def __init__(self, name: str):
        ReferencesV2Command.__init__(self, name, BibliographicReference)


class GeographicReferencesCommand(ReferencesV2Command):
    def __init__(self, name: str):
        ReferencesV2Command.__init__(self, name, GeographicReference)



from collections import OrderedDict

from openpyxl.worksheet.worksheet import Worksheet

from nexinfosys import AreaTupleType, IssuesLabelContentTripleType
from nexinfosys.command_generators import parser_field_parsers, Issue, IssueLocation, IType
from nexinfosys.command_generators.parser_field_parsers import simple_ident
from nexinfosys.common.helper import obtain_dataset_metadata, create_dictionary, strcmp
from nexinfosys.model_services import get_case_study_registry_objects


# TODO Currently is just a copy of "parse_etl_external_dataset_command" function
# TODO It has two new parameters: "InputDataset" and "AvailableAtDateTime"
# TODO Time dimension can be specified as "Time" or as "StartTime" "EndTime"
# TODO Result parameter column also change a bit
# TODO For a reference of fields, see "DatasetQry" command in "MuSIASEM case study commands" Google Spreadsheet

def parse_dataset_qry_command(sh: Worksheet, area: AreaTupleType, name, state) -> IssuesLabelContentTripleType:
    """
    Check that the syntax of the input spreadsheet is correct
    Return the analysis in JSON compatible format, for execution

    :param sh:   Input worksheet
    :param area: Area of the input worksheet to be analysed
    :return:     The command in a dict-list object (JSON ready)
    """
    def obtain_column(cn, r1, r2):
        """
        Obtain a list with the values of a column, in the range of rows [r1, r2)

        :param cn: Column number
        :param r1: Starting row
        :param r2: End+1 row
        :return: list with the cell values
        """
        lst = []
        for row in range(r1, r2):
            value = sh.cell(row=row, column=cn).value
            if value is None:
                continue
            if isinstance(value, str):
                lst.append(value.strip())
            else:
                lst.append(value)
        return lst

    issues = []
    # Global variables (at parse time they may not be defined, so process carefully...)
    glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)

    # Look for the name of the input Dataset
    dataset_name = None
    available_at_datetime = None
    for c in range(area[2], area[3]):
        col_name = sh.cell(row=1, column=c).value
        if not col_name:
            continue
        if col_name.lower().strip() in ["inputdataset"]:
            lst = obtain_column(c, area[0]+1, area[1])
            for v in lst:
                if v:
                    dataset_name = v
                    break  # Stop on first definition
        elif col_name.lower().strip() in ["availableatdatetime"]:
            lst = obtain_column(c, area[0]+1, area[1])
            for v in lst:
                if v:
                    available_at_datetime = v
                    break  # Stop on first definition

    if dataset_name is None:
        issues.append(Issue(itype=IType.ERROR,
                            description=f"The name of the input dataset must be specified under column 'InputDataset'. Skipping {name} command",
                            location=IssueLocation(sheet_name=name, row=None, column=None)))
        return issues, None, None

    # Obtain the source
    from nexinfosys.ie_imports.data_source_manager import DataSourceManager
    source = DataSourceManager.obtain_dataset_source(dataset_name, datasets)
    # Obtain metadata
    dims, attrs, meas = obtain_dataset_metadata(dataset_name, source, datasets)
    # Load all code lists in a temporary dictionary of sets
    # Also check if there is a TIME dimension in the dataset
    cl = create_dictionary()
    we_have_time = False
    for d in dims:
        if dims[d].code_list:
            cl[d] = create_dictionary(data={k: None for k in dims[d].code_list.keys()})  # Attach the code list
        else:
            cl[d] = None  # No code list (TIME_PERIOD for instance)
        if dims[d].istime:
            we_have_time = True

    # Add matching mappings as more dimensions
    for m in mappings:
        if strcmp(mappings[m].source, source) and \
                strcmp(mappings[m].dataset, dataset_name) and \
                mappings[m].origin in dims:
            # Add a dictionary entry for the new dimension, add also the codes present in the map
            # tmp = [to["d"] for o in mappings[m].map for to in o["to"] if to["d"]]
            tmp = create_dictionary(data={to["d"]: None for o in mappings[m].map for to in o["to"] if to["d"]})
            cl[mappings[m].destination] = tmp  # [t[1] for t in mappings[m].map]

    # Scan columns for Dimensions, Measures and Aggregation.
    # Pivot Table is a Visualization, so now it is not in the command, there will be a command aside.

    # TODO The result COULD be an automatic BI cube (with a separate field)
    # TODO - Write into a set of tables in Mondrian
    # TODO - Generate Schema for Mondrian
    # TODO - Write the Schema for Mondrian

    out_dims = []

    out_measures = OrderedDict()
    for r in range(area[0]+1, area[1]+1):
        out_measures[r] = dict(measure=None, agg_func=None, measure_as=None)

    filter_ = {}  # Cannot use "create_dictionary()" because CaseInsensitiveDict is NOT serializable (which is a requirement)
    result_name = None  # By default, no name for the result. It will be dynamically obtained
    measure_names_column = None
    aggregations_column = None
    for c in range(area[2], area[3]):  # Each column
        col_name = sh.cell(row=1, column=c).value
        if not col_name:
            continue
        if col_name.lower().strip() in ["resultdimensions", "dimensions"]:  # "GROUP BY"
            lst = obtain_column(c, area[0] + 1, area[1])
            for r, d in enumerate(lst):
                if not d:
                    continue
                if d not in cl:
                    issues.append(Issue(itype=IType.ERROR,
                                        description="The dimension specified for output, '"+d+"' is neither a dataset dimension nor a mapped dimension. ["+', '.join([d2 for d2 in cl])+"]",
                                        location=IssueLocation(sheet_name=name, row=r + 1, column=c + 1)))
                else:
                    out_dims.append(d)
        elif col_name.lower().strip() in ["resultmeasures", "measures"]:  # "SELECT"
            measure_names_column = c
            lst = obtain_column(c, area[0] + 1, area[1])
            # Check for measures
            # TODO (and attributes?)
            for r, m in enumerate(lst):
                if not m:
                    continue
                if m not in meas:
                    issues.append(Issue(itype=IType.ERROR,
                                        description="The specified measure, '"+m+"' is not a measure available in the dataset. ["+', '.join([m2["measure"] for m2 in out_measures.values])+"]",
                                        location=IssueLocation(sheet_name=name, row=r + 1, column=c + 1)))
                else:
                    out_measures[r+area[0]+1]["measure"] = m
        elif col_name.lower().strip() in ["resultmeasuresaggregation", "resultmeasuresaggregator", "aggregation"]:  # "SELECT AGGREGATORS"
            aggregations_column = c
            lst = obtain_column(c, area[0] + 1, area[1])
            for r, f in enumerate(lst):
                if not f:
                    continue

                if f.lower() not in ["sum", "avg", "count", "sumna", "countav", "avgna", "pctna"]:
                    issues.append(Issue(itype=IType.ERROR,
                                        description="The specified aggregation function, '"+f+"' is not one of the supported ones: 'sum', 'avg', 'count', 'sumna', 'avgna', 'countav', 'pctna'",
                                        location=IssueLocation(sheet_name=name, row=r + 1, column=c + 1)))
                else:
                    out_measures[r+area[0]+1]["agg_func"] = f
        elif col_name.lower().strip() in ["resultmeasurename", "resultmeasuresnames", "resultmeasuresas", "measuresas"]:  # "AS <name>"
            lst = obtain_column(c, area[0] + 1, area[1])
            for r, m in enumerate(lst):
                out_measures[r+area[0]+1]["measure_as"] = m
        elif col_name in cl:  # A dimension -> "WHERE"
            # Check codes, and add them to the "filter"
            lst = obtain_column(c, area[0] + 1, area[1])
            for r, cd in enumerate(lst):
                if not cd:
                    continue
                if str(cd) not in cl[col_name]:
                    issues.append(Issue(itype=IType.ERROR,
                                        description="The code '"+cd+"' is not present in the codes declared for dimension '"+col_name+"'. Please, check them.",
                                        location=IssueLocation(sheet_name=name, row=r + 1, column=c + 1)))
                else:
                    if col_name not in filter_:
                        lst2 = []
                        filter_[col_name] = lst2
                    else:
                        lst2 = filter_[col_name]
                    lst2.append(cd)
        elif we_have_time and col_name.lower() in ["startperiod", "starttime", "endperiod", "endtime"]:  # SPECIAL "WHERE" FOR TIME
            # TODO Instead, should use a single column, "Time", using the interval syntax of the Time column in the Data Input command
            # Interval of time periods
            lst = obtain_column(c, area[0] + 1, area[1])
            if len(lst) > 0:
                if col_name.lower() == "starttime":
                    col_name = "StartPeriod"
                elif col_name.lower() == "endtime":
                    col_name = "EndPeriod"
                filter_[col_name] = lst[0]  # In this case it is not a list, but a number or string !!!!
        elif col_name.lower() in ["outputdatasetname", "outputdataset", "result_name", "result name", "resultname"]:
            lst = obtain_column(c, area[0] + 1, area[1])
            if len(lst) > 0:
                result_name = lst[0]
                try:
                    parser_field_parsers.string_to_ast(simple_ident, result_name)
                except:
                    issues.append(Issue(itype=IType.ERROR,
                                        description="Column '" + col_name + "' has an invalid dataset name '" + result_name + "'",
                                        location=IssueLocation(sheet_name=name, row=2, column=c + 1)))

    # If more than one agg function defined -> all must be defined
    # If no agg func defined -> assume AVG
    # If agg func defined only in first row -> extend to other columns
    agg_funcs = [v["agg_func"] for v in out_measures.values() if v["agg_func"]]
    if len(agg_funcs) > 1:
        first_agg_func = None
    elif len(agg_funcs) == 0:
        issues.append(Issue(itype=IType.WARNING,
                            description="No aggregation function specified. Assuming 'average'",
                            location=IssueLocation(sheet_name=name, row=1, column=aggregations_column)))
        first_agg_func = "avg"
    else:  # One aggregation function
        first_agg_func = out_measures[area[0]+1]["agg_func"]
        if not first_agg_func:
            issues.append(Issue(itype=IType.ERROR,
                                description="The aggregation function must be defined in the first row",
                                location=IssueLocation(sheet_name=name, row=1, column=aggregations_column)))

    if first_agg_func:
        for v in out_measures.values():
            if v.get("measure", None):
                v["agg_func"] = first_agg_func

    # Uniform rows, with the three values defined: measure, aggregation function and "measure as"
    for r, v in out_measures.items():
        measure = v.get("measure", None)
        agg_func = v.get("agg_func", None)
        measure_as = v.get("measure_as", None)
        if measure and not agg_func or not measure and agg_func:
            issues.append(Issue(itype=IType.ERROR,
                                description="Each measure must be associated with an aggregation function",
                                location=IssueLocation(sheet_name=name, row=r, column=measure_names_column)))
        elif measure and not measure_as:
            v["measure_as"] = measure + "_" + agg_func

    measures = [v["measure"] for v in out_measures.values() if v["measure"]]
    measures_as = [v["measure_as"] for v in out_measures.values() if v["measure_as"]]
    agg_funcs = [v["agg_func"] for v in out_measures.values() if v["agg_func"]]

    if len(measures) == 0:
        issues.append(Issue(itype=IType.ERROR,
                            description="At least one measure should be specified",
                            location=IssueLocation(sheet_name=name, row=1, column=measure_names_column)))

    # measures != agg_funcs && len(agg_funcs) == 1 --> OK
    if len(measures) != len(agg_funcs) and len(agg_funcs) != 1:
        issues.append(Issue(itype=IType.ERROR,
                            description="There must be one aggregation function (used for all measures) or one aggregation per measure",
                            location=IssueLocation(sheet_name=name, row=1, column=aggregations_column)))

    if not result_name:
        result_name = source + "_" + dataset_name
        issues.append(Issue(itype=IType.WARNING,
                            description="No result name specified. Assuming '"+result_name+"'",
                            location=IssueLocation(sheet_name=name, row=2, column=c + 1)))

    content = {"dataset_source": source,
               "dataset_name": dataset_name,
               "dataset_datetime": available_at_datetime,
               "where": filter_,
               "dimensions": [d for d in dims],
               "group_by": out_dims,
               "measures": measures,
               "agg_funcs": agg_funcs,
               "measures_as": measures_as,
               "result_name": result_name
               }
    return issues, None, content




import json

from nexinfosys.model_services import IExecutableCommand
from nexinfosys import data_source_manager as dsm


class GetDatasetQueyWorksheetCommand(IExecutableCommand):
    """
    Obtain a
    """
    def __init__(self, name: str):
        self._name = name
        self._content = None

    def execute(self, state: "State"):
        """

        """
        # The data source manager is "dsm"


        state.set(self._var_name, self._description)
        return None, None

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the metadata dictionary
        return self._content

    def json_deserialize(self, json_input):
        # TODO Read and check keys validity
        issues = []
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)
        return issues




import os
import tempfile
from typing import List
import pandas as pd
import sdmx
import pandas_datareader.data as web
import requests
import requests_cache

from nexinfosys.common.helper import Memoize2
from nexinfosys.ie_imports.data_source_manager import IDataSourceManager, filter_dataset_into_dataframe
from nexinfosys.models.statistical_datasets import DataSource, Database, Dataset, Dimension, CodeList, CodeImmutable


class OECD(IDataSourceManager):
    def __init__(self):
        d = {"backend": "sqlite", "include_get_headers": True, "cache_name": tempfile.gettempdir() + "/oecd_bulk_datasets"}
        requests_cache.install_cache(**d)

    def get_name(self) -> str:
        """ Source name """
        return self.get_datasource().name

    def get_datasource(self) -> DataSource:
        """ Data source """
        src = DataSource()
        src.name = "OECD"
        src.description = "OECD is group of 34 democracies with market economy plus 70 economies where growth, prosperity and sustainable development is promoted"
        return src

    def get_databases(self) -> List[Database]:
        """ List of databases in the data source """
        db = Database()
        db.code = ""
        db.description = "OECD provides all Datasets in a single database"
        return [db]

    @Memoize2
    def get_datasets(self, database=None) -> list:
        """ List of datasets in a database, or in all the datasource (if database==None)
            Return a list of tuples (database, dataset)
        """
        from lxml import etree
        # List of datasets, containing three columns: ID, description, ID
        lst = []
        # TODO The third value should be a URN, but it is not available
        xml = requests.get("http://stats.oecd.org/RestSDMX/sdmx.ashx/GetKeyFamily/all")
        root = etree.fromstring(xml.content.decode("utf-8"))
        ids = root.xpath("//*[@agencyID='OECD']/@id")
        descs = root.xpath("//*[@agencyID='OECD']/*[@xml:lang='en']/text()")
        for desc, id_ in zip(descs, ids):
            lst.append((id_, desc, id_))

        return lst

    def get_dataset_structure(self, database, dataset) -> Dataset:
        """ Obtain the structure of a dataset: concepts, dimensions, attributes and measures """
        # Retrieve and save to file
        xml = requests.get("http://stats.oecd.org/restsdmx/sdmx.ashx/GetDataStructure/"+dataset)
        xml = xml.content.decode("utf-8")
        _, fname = tempfile.mkstemp(".oecd_dsd.xml", text=True)
        file = open(fname, "w")
        file.writelines(xml)
        file.close()

        dsd = sdmx.dsd_reader(fname)
        kfam = dsd.key_families()[0]

        # SDMXConcept = collections.namedtuple('Concept', 'type name istime description code_list')
        # DataSource <- Database <- DATASET <- Dimension(s) (including Measures) <- CodeList
        #                                      |
        #                                      v
        #                                      Concept <- CodeList  (NOT CONSIDERED NOW)
        ds = Dataset()
        ds.code = dataset
        ds.description = kfam.name("en")
        ds.attributes = {}  # Dataset level attributes? (encode them using a dictionary)
        ds.metadata = None  # Metadata for the dataset SDMX (flow, date of production, etc.)
        ds.database = database  # Reference to containing database

        # Code lists (assume flat)
        c_dict = {}
        for cl in dsd.code_lists():
            codes = [CodeImmutable(c.value, c.description("en"), "", []) for c in cl.codes()]
            c_dict[cl.id] = CodeList.construct(cl.id, cl.name("en"), [""], codes=codes)

        # Time dimension name
        if kfam.time_dimension():
            t_dim = kfam.time_dimension().concept_ref()
            t_dim_cl = kfam.time_dimension().code_list_id()
        else:
            t_dim = None

        # Dimensions
        for d in kfam.dimensions():
            dd = Dimension()
            dd.code = d.concept_ref()
            dd.attributes = None
            istime = dd.code == t_dim
            if istime:
                dd.is_time = istime
                dd.code_list = c_dict[t_dim_cl]
            else:
                dd.code_list = c_dict[d.code_list_id()]
            dd.description = dd.code_list.description
            dd.is_measure = False  # DIMENSION
            dd.dataset = ds

        # Measure
        dd = Dimension()
        dd.code = kfam.primary_measure().concept_ref()
        dd.description = dsd.concept(dd.code).name("en")
        dd.attributes = None
        dd.is_time = False
        dd.is_measure = True
        dd.dataset = ds

        return ds

    def etl_full_database(self, database=None, update=False):
        """ If bulk download is supported, refresh full database """
        pass

    def etl_dataset(self, dataset, update=False) -> str:
        """
        Download a OECD bulk file

        :param url:
        :param local_filename:
        :param update:
        :return: String with full file name
        """
        fname = tempfile.gettempdir() + "/oecd_" + dataset + '.pickled_dataframe'

        if os.path.isfile(fname):
            if not update:
                return fname
            else:
                os.remove(fname)

        df = web.DataReader(dataset, "oecd")
        df.to_pickle(fname)

        return fname

    def get_dataset_filtered(self, dataset: str, dataset_params: dict) -> Dataset:
        """ This method has to consider the last dataset download, to re"""

        # Read OECD dataset structure
        ds = self.get_dataset_structure(None, dataset)

        # Read full OECD dataset into a Dataframe
        dataframe_fn = tempfile.gettempdir() + "/oecd_" + dataset + ".bin2"
        df = None
        if os.path.isfile(dataframe_fn):
            df = pd.read_parquet(dataframe_fn)
            # df = pd.read_msgpack(dataframe_fn)

        if df is None:
            # Read the unprocessed dataset
            fname = self.etl_dataset(dataset, update=False)
            df = pd.read_pickle(fname)

            # Process DataFrame to make it compatible with "filter_dataset_into_dataframe"
            df = df.transpose()
            # Date to Year dictionary
            df.columns = pd.DatetimeIndex(df.columns.values, name=df.columns.name).year
            # t_dict = {d: y for d, y in zip(df.columns.values, pd.DatetimeIndex(df.columns.values).year)}
            # df.replace({"Year": t_dict}, inplace=True)
            # Time dimension is incorporated with
            ser = df.stack().sort_index()
            ser.name = "value"
            # By removing the index, the series is again a pd.DataFrame
            df = ser.reset_index()

            # Convert descriptions to codes
            recodes = {}
            for d in ds.dimensions:
                if d.code_list:
                    codes = d.code_list.levels[0].codes
                    if len(codes) > 0:
                        recodes[d.description] = {c.description: c.code.lower() for c in codes}
                else:
                    pass

            df.replace(recodes, inplace=True)

            df.set_index([c for c in df.columns[:-1]], inplace=True)
            # Save it
            df.to_parquet(dataframe_fn)
            # df.to_msgpack(dataframe_fn)

        # Filter it using generic Pandas filtering capabilities
        ds.data = filter_dataset_into_dataframe(df, dataset_params, dataset)

        return ds

    def get_refresh_policy(self):  # Refresh frequency for list of databases, list of datasets, and dataset
        pass


# o = OECD()
# filter_ = create_dictionary(data=dict(country=["AUT", "AUS"]))
# df = o.get_dataset_filtered("TUD", filter_)
# print(df)



# -*- coding: utf-8 -*-
"""
This generates and calls the api

It's worth writing this so that a database is updated - maybe csv based for
each commodity code

@author: Eoin O'Keeffe

@date 08/08/2014

"""
import requests
import pandas as pd
from os.path import join
import os
import time
import unicodedata
import numpy
import datetime
import sys
from os.path import isfile
import logging


class ComtradeApi:
    _source_folder = ""
    _url = 'http://comtrade.un.org/api/get?'
    _ctry_codes = []
    _ctry_alt_names = []
    _max_partners = 5
    _max_years = 5
    _last_call_time = 0
    _working_df = []
    calls_in_hour = 0
    first_call = datetime.datetime.now()
    max_calls = 95
    """
    Pure api call, no updating of a database


    """

    def __init__(self, ctry_codes_path="UN Comtrade Country List.csv", fld=""):
        self._source_folder = fld
        # load the country codes
        self._ctry_codes = pd.read_csv(
            join(fld, ctry_codes_path), keep_default_na=False, encoding="ISO-8859-1")
        # Substitute "End Valid Year" values "Now" with current year
        import datetime
        year = datetime.datetime.now().year
        self._ctry_codes.loc[self._ctry_codes["End Valid Year"] == "Now", "End Valid Year"] = year
        self._ctry_codes['End Valid Year'] = self._ctry_codes['End Valid Year'].astype("int32")
        self._ctry_codes = self._ctry_codes.ix[self._ctry_codes['End Valid Year'] > 2012]
        # check to make sure there are correct fields
        if ('ISO2-digit Alpha' not in self._ctry_codes.columns) | \
                ('ISO3-digit Alpha' not in self._ctry_codes.columns):
            logging.warning("----------------------------")
            logging.warning("                 ")
            logging.warning("Check format of %s file" % ctry_codes_path)
            logging.warning("It appears to be missing fields")
            logging.warning("                 ")
            logging.warning("--------------------------")
            return
        # Remove NES and other areas
        # check the field name for the country code
        if "Country Code" in self._ctry_codes.columns:
            self._ctry_codes['ctyCode'] = self._ctry_codes['Country Code']
        # World
        self._ctry_codes = self._ctry_codes.ix[self._ctry_codes['ctyCode'] != 0]
        # EU-27
        self._ctry_codes = self._ctry_codes.ix[self._ctry_codes['ctyCode'] != 97]
        # LAIA NES
        self._ctry_codes = self._ctry_codes.ix[self._ctry_codes['ctyCode'] != 473]
        # Oceania NES
        self._ctry_codes = self._ctry_codes.ix[self._ctry_codes['ctyCode'] != 527]
        # Europe NES
        self._ctry_codes = self._ctry_codes.ix[self._ctry_codes['ctyCode'] != 568]
        # Other Africa NES
        self._ctry_codes = self._ctry_codes.ix[self._ctry_codes['ctyCode'] != 577]
        # Bunkers
        self._ctry_codes = self._ctry_codes.ix[self._ctry_codes['ctyCode'] != 837]
        # Free Zones
        self._ctry_codes = self._ctry_codes.ix[self._ctry_codes['ctyCode'] != 838]
        # Special Caegories
        self._ctry_codes = self._ctry_codes.ix[self._ctry_codes['ctyCode'] != 839]
        # Areas NES
        self._ctry_codes = self._ctry_codes.ix[self._ctry_codes['ctyCode'] != 899]
        # Neutral Zone
        self._ctry_codes = self._ctry_codes.ix[self._ctry_codes['ctyCode'] != 536]
        # North America and Central America, nes
        self._ctry_codes = self._ctry_codes.ix[self._ctry_codes['ctyCode'] != 637]
        # Finally remove where the iso codes are not available
        self._ctry_codes = self._ctry_codes.ix[pd.isnull(self._ctry_codes['ISO2-digit Alpha']) == False]
        self._ctry_codes = self._ctry_codes.ix[pd.isnull(self._ctry_codes['ISO3-digit Alpha']) == False]

        # now the country code alternative names
        # if this doesn't exist then create it
        if not isfile(join(fld, "country_alternative_names.csv")):
            # doesn't exist so create it
            self._ctry_alt_names = pd.DataFrame({'ISO2-digit Alpha': ['KR'],
                                                 'Aka': ['South Korea']})
            # write this to file
            self._ctry_alt_names.to_csv(
                join(fld, "country_alternative_names.csv"), index=None)
        self._ctry_alt_names = pd.read_csv(join(fld, "country_alternative_names.csv"),
                                           keep_default_na=False)

        saved_queries_fname = join(self._source_folder, "saved_queries.csv")
        if (os.path.isfile(saved_queries_fname)):
            self._saved_queries = pd.read_csv(saved_queries_fname)
        else:
            self._saved_queries = pd.DataFrame({'id': [], 'querystring': [],
                                                'comcode': [], 'year': [], 'freq': []})
        self._saved_queries.comcode = self._saved_queries.comcode.astype(
            pd.np.str)
        self._saved_queries.freq = self._saved_queries.freq.astype(pd.np.str)
        self._saved_queries.querystring = self._saved_queries.querystring.astype(
            pd.np.str)

    # ---------------------------------
    #
    # Call this to get data
    #
    # @param comcodes - array/list of HS country codes as strings
    # @param reporter - array/list of comtrade countries as list or use 'all' for
    # all countries
    # @param partner - same as above
    # @param years -array of years as integers
    # @param freq - 'A' or 'M' for year or monthly data respectively
    # @param rg='1' for imports and '2' for exports - only tested for imports
    # @param fmt='json' - default
    # @param rowmax=50000. No reason to change this. This is the max value.
    # ----------------------------------
    def getComtradeData(self, comcodes=['1201'], reporter=['all'], partner=['all'],
                        years=[2012], freq='A', rg=['1'], fmt='json', rowmax=50000):

        # build string
        s = self._url + "type=C"
        s = "%s&freq=%s" % (s, freq)
        s = "%s&px=HS" % s
        if freq == 'A':
            s = "%s&ps=%s" % (s, "%2C".join([str(yr) for yr in years]))
        else:
            mons = [["%s%02d" % (yr, mon) for mon in range(1, 13)]
                    for yr in years]
            s = "%s&ps=%s" % (s, "%2C".join(mons[0]))

        s = "%s&r=%s" % (s, "%2C".join(reporter))

        # For large queries, it's possible we've stored the results
        if reporter[0] == "all" and partner[0] == "all":
            # Check if query has already been done
            df_base = []
            retain = []
            load_files = []
            for yr in years:
                haveit = True
                for com in comcodes:
                    if len(self._saved_queries.ix[(self._saved_queries.comcode == com) &
                                                  (self._saved_queries.year == yr) & (
                                                          self._saved_queries.freq == freq)]) == 0:
                        logging.info(
                            "Can't find %d and %s from local database. These will be retrieved from the Comtrade API." % (
                            yr, com))
                        retain.append(yr)
                        haveit = False
                        break
                if haveit == True:
                    # print "we have it"
                    for com in comcodes:
                        fname = self._saved_queries.id.ix[(self._saved_queries.comcode == com) &
                                                          (self._saved_queries.year == yr) & (
                                                                      self._saved_queries.freq == freq)].values[0]
                        # only load it if fname if >0
                        if (fname > -1):
                            load_files.append(fname)
                        else:
                            logging.info("{} for year {} and freq {} not available in comtrade. Fname: {}".format(
                                com, yr, freq, fname))
            load_files = numpy.unique(load_files)
            for fname in load_files:
                fname = join(self._source_folder, "%d.csv" % fname)
                df_new = pd.read_csv(fname)
                # only keep current year
                if len(df_base) == 0:
                    df_base = df_new
                else:
                    df_base = df_base.append(df_new)
                    df_base.drop_duplicates(inplace=True)
                    # Remove the year from years
                    # for_removal.append(years.index(yr))

                # Now remove years and coms outside the requested
                # logging.info(df_base.shape)
                # df_base=df_base.ix[df_base.yr.apply(lambda x: x in years)].copy()
                # logging.info(df_base.shape)
                # df_base=df_base.ix[df_base.cmdCode.apply(lambda x: x in [int(x) for x in comcodes])].copy()
                # logging.info(df_base.shape)

            filter_years = years
            years = retain
            # logging.debug(years)
            if (len(years) == 0):
                # print "filled up so jumping out"
                # jump out
                df_base.drop_duplicates(inplace=True)
                # Only return wanted commodity codes

                df_base = df_base.ix[df_base.cmdCode.isin(
                    [int(numeric_string) for numeric_string in comcodes])]

                if freq == 'A':
                    # logging.debug(df_base.head())
                    df_base = df_base.ix[df_base.period.isin(filter_years)]
                    df_base['date'] = [datetime.datetime(
                        x, 1, 1) for x in df_base.period]
                if freq == 'M':
                    df_base['year'] = [int(str(x)[:4])
                                       for x in df_base.period.values]
                    df_base = df_base.ix[df_base.year.isin(filter_years)]
                    df_base['date'] = [datetime.datetime(
                        int(str(x)[:4]), int(str(x)[4:]), 1) for x in df_base.period]
                    df_base.drop('year', inplace=True, axis=1)
                # logging.debug(filter_years)
                # df_base=df_base.ix[df_base.period.isin(years)]
                # reset_index
                df_base.reset_index(drop=True, inplace=True)
                return df_base
            df = []
            # logging.debug(df_base.head())
            # can't pass both as all, iterate through country codes and add each
            # Need to split into calls of
            for start_year in range(0, len(years), self._max_years):
                df_cur_years = []
                for start_val in range(0, len(self._ctry_codes.ctyCode), self._max_partners):
                    end_val = min([start_val + self._max_partners,
                                   len(self._ctry_codes.ctyCode)])
                    end_year = min([start_year + self._max_years, len(years)])
                    # logging.info("%d,%d"%(start_year,end_year))
                    sub_partners = self._ctry_codes.ctyCode.values[
                                   start_val:end_val]
                    sub_partners = [str(com) for com in sub_partners]
                    logging.debug("Running {} of {} subqueries for year(s) ".format(
                        start_val, len(self._ctry_codes.ctyCode), years[start_year:end_year]))
                    # print sub_partners
                    new_data = self.getComtradeData(comcodes=comcodes, reporter=reporter, partner=sub_partners,
                                                    years=years[start_year:end_year], freq=freq, rg=rg, fmt=fmt,
                                                    rowmax=rowmax)
                    if len(df_cur_years) == 0:
                        df_cur_years = new_data
                    else:
                        # pass
                        try:
                            # print "appending..."
                            df_cur_years = df_cur_years.append(new_data)
                        except e:
                            logging.warning(
                                "Error trying to append new service call data")
                            exception_name, exception_value = sys.exc_info()[
                                                              :2]
                            raise  # or don't -- it's up to you
                            # print "Appending..."

                        # if start_val>40:
                        # break
                # Now save this call to file
                # save this query to disk
                id = 0
                if len(df_cur_years) == 0:
                    id = -1
                elif (len(self._saved_queries.id) == 0):
                    id = 0
                else:
                    id = max(self._saved_queries.id.values) + 1
                # if len(df)>0:
                logging.info(
                    "Writing {}/{} to file with id:{}".format(comcodes, years[start_year:end_year], id))
                for yr in years[start_year:end_year]:
                    for com in comcodes:
                        df_new = pd.DataFrame({'id': [id], 'querystring': [s],
                                               'comcode': [com], 'year': [yr], 'freq': [freq]})
                        self._saved_queries = self._saved_queries.append(
                            df_new)
                self._saved_queries.to_csv(join(self._source_folder,
                                                "saved_queries.csv"), index=None)
                if id > -1:
                    # unencode from unicode
                    df_cur_years.rtTitle = df_cur_years.rtTitle.apply(lambda x:
                                                                      unicodedata.normalize('NFKD', x).encode('ascii',
                                                                                                              'ignore'))
                    df_cur_years.ptTitle = df_cur_years.ptTitle.apply(lambda x:
                                                                      unicodedata.normalize('NFKD', x).encode('ascii',
                                                                                                              'ignore'))
                    # self._working_df=df_cur_years
                    # write to file

                    # self._working_df=df_cur_years
                    df_cur_years.to_csv(
                        join(self._source_folder, "%d.csv" % id), index=None)

                    # Now merge this with df
                    if len(df) == 0:
                        df = df_cur_years
                    else:
                        try:
                            # print "appending..."
                            df = df.append(df_cur_years)
                        except e:
                            logging.warning("Error trying to append. Years: {}. Com: {}".format(
                                years[start_year:end_year], comcodes))
                            exception_name, exception_value = sys.exc_info()[
                                                              :2]
                            raise
            logging.debug("Have the merged dataset")

            # print df.head()

            if len(df_base) > 0:
                # print df_base.head()
                df = df.append(df_base)
                # keep unique rows
                # print df.head()
                if len(df) > 0:
                    df.drop_duplicates(inplace=True)
            # Only return wanted commodity codes
            if len(df) > 0:
                # ensure cmdCode is an integer
                df.cmdCode = df.cmdCode.astype(pd.np.int)
                df.period = df.period.astype(pd.np.int)
                # print df.head()
                # print comcodes
                # print filter_years
                # print df.dtypes
                df = df.ix[df.cmdCode.isin(
                    [int(numeric_string) for numeric_string in comcodes])]
                # if freq=='A':
                df = df.ix[df.yr.isin(filter_years)]
                # print df.head()
            return df
        else:
            s = "%s&p=%s" % (s, "%2C".join(partner))
        s = "%s&rg=%s" % (s, "%2C".join(rg))
        s = "%s&cc=%s" % (s, "%2C".join(comcodes))
        s = "%s&fmt=%s" % (s, fmt)
        s = "%s&max=%d" % (s, rowmax)
        s = "%s&head=M" % (s)

        # first we need to figure out if we can do the call - we may have
        # exceeded our allowable calls
        if (ComtradeApi.first_call < datetime.datetime.now() + datetime.timedelta(hours=-1)):
            # we've waiting long enough
            ComtradeApi.first_call = datetime.datetime.now()
            ComtradeApi.calls_in_hour = 0
        else:
            # we're within time, so check how many we have left
            if (ComtradeApi.calls_in_hour + 1 < ComtradeApi.max_calls):
                logging.info("you've made %d calls this hour, %d left" %
                             (ComtradeApi.calls_in_hour + 1,
                              ComtradeApi.max_calls - ComtradeApi.calls_in_hour - 1))
                ComtradeApi.calls_in_hour += 1
            else:
                # Need to wait for an hour

                waiting_time = datetime.timedelta(hours=1.05)
                logging.warning("You've exceeded the max calls in an hour, so now you'll have to wait until {}".format(
                    datetime.datetime.now() + waiting_time))
                sys.stdout.flush()
                time.sleep(waiting_time.total_seconds())
                logging.debug("Stepping out of waiting. Time: {}".format(
                    datetime.datetime.now()))
                ComtradeApi.first_call = datetime.datetime.now()
                ComtradeApi.calls_in_hour = 0
        # print "Doing an api call"
        logging.info(s)
        waiting_time = round(time.time() * 1000) - self._last_call_time
        if (waiting_time < 1000):
            # print waiting_time
            # print("Sleeping for %.0f"%(1000-waiting_time))
            time.sleep(float(1000 - waiting_time) / 1000)
        r = requests.get(r'%s' % (s))
        self._last_call_time = round(time.time() * 1000.0)
        # print self._last_call_time
        try:
            data = r.json()
        except:
            logging.warning("No json object to be parsed")
            logging.warning(
                "Trying running the string below in your browser -is there an error?")
            logging.warning("return string: {}".format(r))
            logging.warning("%s" % s)
            exception_name, exception_value = sys.exc_info()[:2]
            logging.warning("{} {}".format(exception_name, exception_value))
            raise

            return []
        df = pd.DataFrame(data['dataset'])
        logging.info("Returned rows: %d" % df.shape[0])

        return df


api = ComtradeApi()
df = api.getComtradeData(rg=['2'])
print(df.head)



from openpyxl.worksheet.worksheet import Worksheet

from nexinfosys import metadata_fields, AreaTupleType, IssuesLabelContentTripleType
from nexinfosys.common.helper import create_dictionary


def parse_metadata_command(sh: Worksheet, area: AreaTupleType, name: str = None) -> IssuesLabelContentTripleType:
    """
    Most "parse" methods are mostly syntactic (as opposed to semantic). They do not check existence of names.
    But in this case, the valid field names are fixed beforehand, so they are checked at this time.
    Some of the fields will be controlled also, according to some

    :param sh: Input worksheet
    :param area: Tuple (top, bottom, left, right) representing the rectangular area of the input worksheet where the
    command is present
    :return: list of issues (issue_type, message), command label, command content
    """
    some_error = False
    issues = []
    controlled = create_dictionary()
    mandatory = create_dictionary()
    keys = create_dictionary()
    for t in metadata_fields:
        controlled[t[4]] = t[3]
        mandatory[t[4]] = t[2]
        keys[t[0]] = t[4]

    # Scan the sheet, the first column must be one of the keys of "k_list", following
    # columns can contain repeating values

    # Map key to a list of values
    content = {}  # Dictionary of lists, one per metadata key
    for r in range(area[0], area[1]):
        label = sh.cell(row=r, column=area[2]).value
        if label in keys:
            key = keys[label]
            for c in range(area[2]+1, area[3]):
                value = sh.cell(row=r, column=c).value
                if value:
                    value = str(value).strip()
                    if controlled[key]:
                        # Control "value" if the field is controllable
                        cl = {"dimensions": ["water", "energy", "food", "land", "climate"],
                              "subject_topic_keywords": None,
                              "geographical_level": ["local", "regional", "region", "country", "europe", "global", "sectoral", "sector"],
                              "geographical_situation": None,  # TODO Read the list of all geographical regions (A long list!!)
                              "restriction_level": ["internal", "confidential", "public"],
                              "language": None,  # TODO Read the list of ALL languages (or just "English"??)
                              }
                        if cl[key] and value.lower() not in cl[key]:
                            issues.append((3, "The key '"+key+"' should be one of: "+",".join(cl[key])))

                    if key not in content:
                        content[key] = []
                    content[key].append(value)
        else:
            issues.append((2, "Row "+str(r)+": unknown metadata label '"+label+"'"))

    for key in keys.values():
        if mandatory[key] and key not in content:
            some_error = True
            issues.append((3, "The value '"+key+"' is mandatory in the definition of the metadata"))

    return issues, None, content
    # else:
    #     if not some_error:
    #         cmd = MetadataCommand(None)
    #         cmd.json_deserialize(content)
    #     else:
    #         cmd = None
    #     return cmd, issues





import sqlalchemy


def drop_pg_database(sa_str, database_name):
    db_connection_string = sa_str
    data_engine = sqlalchemy.create_engine(db_connection_string, echo=False)
    conn = data_engine.connect()
    conn.execute("commit")
    try:
        conn.execute("drop database "+database_name)
    except:
        pass
    conn.close()
    data_engine.dispose()


def create_pg_database_engine(sa_str, database_name, recreate_db=False):
    if recreate_db:
        drop_pg_database(sa_str, database_name)
    db_connection_string = sa_str
    data_engine = sqlalchemy.create_engine(db_connection_string, echo=False)
    conn = data_engine.connect()
    conn.execute("commit")
    try:
        conn.execute("create database "+database_name)
    except:
        pass
    conn.close()
    data_engine.dispose()
    db_connection_string = sa_str+database_name
    return sqlalchemy.create_engine(db_connection_string, echo=False)


def create_monet_database_engine(sa_str, database_name):
    """
    docker run -d --name monetdb -P -p 50000:50000 -v /home/rnebot/DATOS/monetdb:/var/monetdb5/dbfarm monetdb/monetdb-r-docker
    import pymonetdb
    connection = pymonetdb.connect(username="monetdb", password="monetdb", hostname="localhost", database="db")

    MonetDB container
     - Create MonetDB "preparer", using "monet-r-docker" image
     - After container creation, create new DB
       - Implies
    Contenedor MONETDB
     - Crear preparador de instancia Monet, basado en Monet-R
       - Crear+iniciar contenedor
       - Crear BDD
    """
    db_connection_string = sa_str+database_name
    return sqlalchemy.create_engine(db_connection_string, echo=False)
